"""
Factory module
"""

from ...util import Resolver

from .huggingface import HFGeneration
from .litellm import LiteLLM
from .llama import LlamaCpp


class GenerationFactory:
    """
    Methods to create generative models.
    """

    @staticmethod
    def create(path, method, **kwargs):
        """
        Creates a new Generation instance.

        Args:
            path: model path
            method: llm framework
            kwargs: model keyword arguments
        """

        # Derive method
        method = GenerationFactory.method(path, method)

        # LiteLLM generation
        if method == "litellm":
            return LiteLLM(path, **kwargs)

        # llama.cpp generation
        if method == "llama.cpp":
            return LlamaCpp(path, **kwargs)

        # Hugging Face Transformers generation
        if method == "transformers":
            return HFGeneration(path, **kwargs)

        # Resolve custom method
        return GenerationFactory.resolve(path, method, **kwargs)

    @staticmethod
    def method(path, method):
        """
        Get or derives the LLM framework.

        Args:
            path: model path
            method: llm framework

        Return:
            llm framework
        """

        if not method:
            if LiteLLM.ismodel(path):
                method = "litellm"
            elif LlamaCpp.ismodel(path):
                method = "llama.cpp"
            else:
                method = "transformers"

        return method

    @staticmethod
    def resolve(path, method, **kwargs):
        """
        Attempt to resolve a custom LLM framework.

        Args:
            path: model path
            method: llm framework
            kwargs: model keyword arguments

        Returns:
            Generation instance
        """

        try:
            return Resolver()(method)(path, **kwargs)
        except Exception as e:
            raise ImportError(f"Unable to resolve generation framework: '{method}'") from e



"""
Template module
"""

from string import Formatter

from ...util import TemplateFormatter
from .file import Task


class TemplateTask(Task):
    """
    Task that generates text from a template and task inputs. Templates can be used to prepare data for a number of tasks
    including generating large language model (LLM) prompts.
    """

    def register(self, template=None, rules=None, strict=True):
        """
        Read template parameters.

        Args:
            template: prompt template
            rules: parameter rules
            strict: requires all task inputs to be consumed by template, defaults to True
        """

        # pylint: disable=W0201
        # Template text
        self.template = template if template else self.defaulttemplate()

        # Template processing rules
        self.rules = rules if rules else self.defaultrules()

        # Create formatter
        self.formatter = TemplateFormatter() if strict else Formatter()

    def prepare(self, element):
        # Check if element matches any processing rules
        match = self.match(element)
        if match:
            return match

        # Apply template processing, if applicable
        if self.template:
            # Pass dictionary as named prompt template parameters
            if isinstance(element, dict):
                return self.formatter.format(self.template, **element)

            # Pass tuple as prompt template parameters (arg0 - argN)
            if isinstance(element, tuple):
                return self.formatter.format(self.template, **{f"arg{i}": x for i, x in enumerate(element)})

            # Default behavior is to use input as {text} parameter in prompt template
            return self.formatter.format(self.template, text=element)

        # Return original inputs when no prompt provided
        return element

    def defaulttemplate(self):
        """
        Generates a default template for this task. Base method returns None.

        Returns:
            default template
        """

        return None

    def defaultrules(self):
        """
        Generates a default rules for this task. Base method returns an empty dictionary.

        Returns:
            default rules
        """

        return {}

    def match(self, element):
        """
        Check if element matches any processing rules.

        Args:
            element: input element

        Returns:
            matching value if found, None otherwise
        """

        if self.rules and isinstance(element, dict):
            # Check if any rules are matched
            for key, value in self.rules.items():
                if element[key] == value:
                    return element[key]

        return None


class RagTask(TemplateTask):
    """
    Template task that prepares input for a rag pipeline.
    """

    def prepare(self, element):
        # Apply prompt template using all variables except "query" and use output as question
        if isinstance(element, dict):
            # Make a copy without query and run through template
            params = dict(element)
            params.pop("query", None)
            params["text"] = params.pop("question")

            element["question"] = super().prepare(params)
            return element

        # Default mode is to use element text for both query and question
        return {"query": element, "question": super().prepare(element)}



"""
RetrieveTask module
"""

import os
import tempfile

from urllib.request import urlretrieve
from urllib.parse import urlparse

from .url import UrlTask


class RetrieveTask(UrlTask):
    """
    Task that retrieves urls (local or remote) to a local directory.
    """

    def register(self, directory=None, flatten=True):
        """
        Adds retrieve parameters to task.

        Args:
            directory: local directory used to store retrieved files
            flatten: flatten input directory structure, defaults to True
        """

        # pylint: disable=W0201
        # Create default temporary directory if not specified
        if not directory:
            # Save tempdir to prevent content from being deleted until this task is out of scope
            # pylint: disable=R1732
            self.tempdir = tempfile.TemporaryDirectory()
            directory = self.tempdir.name

        # Create output directory if necessary
        os.makedirs(directory, exist_ok=True)

        self.directory = directory
        self.flatten = flatten

    def prepare(self, element):
        # Extract file path from URL
        path = urlparse(element).path

        if self.flatten:
            # Flatten directory structure (default)
            path = os.path.join(self.directory, os.path.basename(path))
        else:
            # Derive output path
            path = os.path.join(self.directory, os.path.normpath(path.lstrip("/")))
            directory = os.path.dirname(path)

            # Create local directory, if necessary
            os.makedirs(directory, exist_ok=True)

        # Retrieve URL
        urlretrieve(element, path)

        # Return new file path
        return path



"""
Application module
"""

import os

from multiprocessing.pool import ThreadPool
from threading import RLock

import yaml

from ..agent import Agent
from ..embeddings import Documents, Embeddings
from ..pipeline import PipelineFactory
from ..workflow import WorkflowFactory


# pylint: disable=R0904
class Application:
    """
    Builds YAML-configured txtai applications.
    """

    @staticmethod
    def read(data):
        """
        Reads a YAML configuration file.

        Args:
            data: input data

        Returns:
            yaml
        """

        if isinstance(data, str):
            if os.path.exists(data):
                # Read yaml from file
                with open(data, "r", encoding="utf-8") as f:
                    # Read configuration
                    return yaml.safe_load(f)

            # Attempt to read yaml from input
            data = yaml.safe_load(data)
            if not isinstance(data, str):
                return data

            # File not found and input is not yaml, raise error
            raise FileNotFoundError(f"Unable to load file '{data}'")

        # Return unmodified
        return data

    def __init__(self, config, loaddata=True):
        """
        Creates an Application instance, which encapsulates embeddings, pipelines and workflows.

        Args:
            config: index configuration
            loaddata: If True (default), load existing index data, if available. Otherwise, only load models.
        """

        # Initialize member variables
        self.config, self.documents, self.embeddings = Application.read(config), None, None

        # Write lock - allows only a single thread to update embeddings
        self.lock = RLock()

        # ThreadPool - runs scheduled workflows
        self.pool = None

        # Create pipelines
        self.createpipelines()

        # Create workflows
        self.createworkflows()

        # Create agents
        self.createagents()

        # Create embeddings index
        self.indexes(loaddata)

    def __del__(self):
        """
        Close threadpool when this object is garbage collected.
        """

        if hasattr(self, "pool") and self.pool:
            self.pool.close()
            self.pool = None

    def createpipelines(self):
        """
        Create pipelines.
        """

        # Pipeline definitions
        self.pipelines = {}

        # Default pipelines
        pipelines = list(PipelineFactory.list().keys())

        # Add custom pipelines
        for key in self.config:
            if "." in key:
                pipelines.append(key)

        # Move dependent pipelines to end of list
        dependent = ["similarity", "extractor", "rag", "reranker"]
        pipelines = sorted(pipelines, key=lambda x: dependent.index(x) + 1 if x in dependent else 0)

        # Create pipelines
        for pipeline in pipelines:
            if pipeline in self.config:
                config = self.config[pipeline] if self.config[pipeline] else {}

                # Add application reference, if requested
                if "application" in config:
                    config["application"] = self

                # Custom pipeline parameters
                if pipeline in ["extractor", "rag"]:
                    if "similarity" not in config:
                        # Add placeholder, will be set to embeddings index once initialized
                        config["similarity"] = None

                    # Resolve reference pipelines
                    if config.get("similarity") in self.pipelines:
                        config["similarity"] = self.pipelines[config["similarity"]]

                    if config.get("path") in self.pipelines:
                        config["path"] = self.pipelines[config["path"]]

                elif pipeline == "similarity" and "path" not in config and "labels" in self.pipelines:
                    config["model"] = self.pipelines["labels"]

                elif pipeline == "reranker":
                    config["embeddings"] = None
                    config["similarity"] = self.pipelines["similarity"]

                self.pipelines[pipeline] = PipelineFactory.create(config, pipeline)

    def createworkflows(self):
        """
        Create workflows.
        """

        # Workflow definitions
        self.workflows = {}

        # Create workflows
        if "workflow" in self.config:
            for workflow, config in self.config["workflow"].items():
                # Create copy of config
                config = config.copy()

                # Resolve callable functions
                config["tasks"] = [self.resolvetask(task) for task in config["tasks"]]

                # Resolve stream functions
                if "stream" in config:
                    config["stream"] = self.resolvetask(config["stream"])

                # Get scheduler config
                schedule = config.pop("schedule", None)

                # Create workflow
                self.workflows[workflow] = WorkflowFactory.create(config, workflow)

                # Schedule job if necessary
                if schedule:
                    # Create pool if necessary
                    if not self.pool:
                        self.pool = ThreadPool()

                    self.pool.apply_async(self.workflows[workflow].schedule, kwds=schedule)

    def createagents(self):
        """
        Create agents.
        """

        # Agent definitions
        self.agents = {}

        # Create agents
        if "agent" in self.config:
            for agent, config in self.config["agent"].items():
                # Create copy of config
                config = config.copy()

                # Resolve LLM
                config["llm"] = self.function("llm")

                # Resolve tools
                for tool in config.get("tools", []):
                    if isinstance(tool, dict) and "target" in tool:
                        tool["target"] = self.function(tool["target"])

                # Create agent
                self.agents[agent] = Agent(**config)

    def indexes(self, loaddata):
        """
        Initialize an embeddings index.

        Args:
            loaddata: If True (default), load existing index data, if available. Otherwise, only load models.
        """

        # Get embeddings configuration
        config = self.config.get("embeddings")
        if config:
            # Resolve application functions in embeddings config
            config = self.resolveconfig(config.copy())

        # Load embeddings index if loaddata and index exists
        if loaddata and Embeddings().exists(self.config.get("path"), self.config.get("cloud")):
            # Initialize empty embeddings
            self.embeddings = Embeddings()

            # Pass path and cloud settings. Set application functions as config overrides.
            self.embeddings.load(
                self.config.get("path"),
                self.config.get("cloud"),
                {key: config[key] for key in ["functions", "transform"] if key in config} if config else None,
            )

        elif "embeddings" in self.config:
            # Create new embeddings with config
            self.embeddings = Embeddings(config)

        # If an extractor pipeline is defined and the similarity attribute is None, set to embeddings index
        for key in ["extractor", "rag"]:
            pipeline = self.pipelines.get(key)
            config = self.config.get(key)

            if pipeline and config is not None and config["similarity"] is None:
                pipeline.similarity = self.embeddings

        # Attach embeddings to reranker
        if "reranker" in self.pipelines:
            self.pipelines["reranker"].embeddings = self.embeddings

    def resolvetask(self, task):
        """
        Resolves callable functions for a task.

        Args:
            task: input task config
        """

        # Check for task shorthand syntax
        task = {"action": task} if isinstance(task, (str, list)) else task

        if "action" in task:
            action = task["action"]
            values = [action] if not isinstance(action, list) else action

            actions = []
            for a in values:
                if a in ["index", "upsert"]:
                    # Add queue action to buffer documents to index
                    actions.append(self.add)

                    # Override and disable unpacking for indexing actions
                    task["unpack"] = False

                    # Add finalize to trigger indexing
                    task["finalize"] = self.upsert if a == "upsert" else self.index
                elif a == "search":
                    actions.append(self.batchsearch)
                elif a == "transform":
                    # Transform vectors
                    actions.append(self.batchtransform)

                    # Override and disable one-to-many transformations
                    task["onetomany"] = False
                else:
                    # Resolve action to callable function
                    actions.append(self.function(a))

            # Save resolved action(s)
            task["action"] = actions[0] if not isinstance(action, list) else actions

        # Resolve initializer
        if "initialize" in task and isinstance(task["initialize"], str):
            task["initialize"] = self.function(task["initialize"])

        # Resolve finalizer
        if "finalize" in task and isinstance(task["finalize"], str):
            task["finalize"] = self.function(task["finalize"])

        return task

    def resolveconfig(self, config):
        """
        Resolves callable functions stored in embeddings configuration.

        Args:
            config: embeddings config

        Returns:
            resolved config
        """

        if "functions" in config:
            # Resolve callable functions
            functions = []
            for fn in config["functions"]:
                original = fn
                try:
                    if isinstance(fn, dict):
                        fn = fn.copy()
                        fn["function"] = self.function(fn["function"])
                    else:
                        fn = self.function(fn)

                # pylint: disable=W0703
                except Exception:
                    # Not a resolvable function, pipeline or workflow - further resolution will happen in embeddings
                    fn = original

                functions.append(fn)

            config["functions"] = functions

        if "transform" in config:
            # Resolve transform function
            config["transform"] = self.function(config["transform"])

        return config

    def function(self, function):
        """
        Get a handle to a callable function.

        Args:
            function: function name

        Returns:
            resolved function
        """

        # Check if function is a pipeline
        if function in self.pipelines:
            return self.pipelines[function]

        # Check if function is a workflow
        if function in self.workflows:
            return self.workflows[function]

        # Attempt to resolve action as a callable function
        return PipelineFactory.create({}, function)

    def search(self, query, limit=10, weights=None, index=None, parameters=None, graph=False):
        """
        Finds documents most similar to the input query. This method will run either an index search
        or an index + database search depending on if a database is available.

        Args:
            query: input query
            limit: maximum results
            weights: hybrid score weights, if applicable
            index: index name, if applicable
            parameters: dict of named parameters to bind to placeholders
            graph: return graph results if True

        Returns:
            list of {id: value, score: value} for index search, list of dict for an index + database search
        """

        if self.embeddings:
            with self.lock:
                results = self.embeddings.search(query, limit, weights, index, parameters, graph)

            # Unpack (id, score) tuple, if necessary. Otherwise, results are dictionaries.
            return results if graph else [{"id": r[0], "score": float(r[1])} if isinstance(r, tuple) else r for r in results]

        return None

    def batchsearch(self, queries, limit=10, weights=None, index=None, parameters=None, graph=False):
        """
        Finds documents most similar to the input queries. This method will run either an index search
        or an index + database search depending on if a database is available.

        Args:
            queries: input queries
            limit: maximum results
            weights: hybrid score weights, if applicable
            index: index name, if applicable
            parameters: list of dicts of named parameters to bind to placeholders
            graph: return graph results if True

        Returns:
            list of {id: value, score: value} per query for index search, list of dict per query for an index + database search
        """

        if self.embeddings:
            with self.lock:
                search = self.embeddings.batchsearch(queries, limit, weights, index, parameters, graph)

            results = []
            for result in search:
                # Unpack (id, score) tuple, if necessary. Otherwise, results are dictionaries.
                results.append(result if graph else [{"id": r[0], "score": float(r[1])} if isinstance(r, tuple) else r for r in result])
            return results

        return None

    def add(self, documents):
        """
        Adds a batch of documents for indexing.

        Args:
            documents: list of {id: value, data: value, tags: value}

        Returns:
            unmodified input documents
        """

        # Raise error if index is not writable
        if not self.config.get("writable"):
            raise ReadOnlyError("Attempting to add documents to a read-only index (writable != True)")

        if self.embeddings:
            with self.lock:
                # Create documents file if not already open
                if not self.documents:
                    self.documents = Documents()

                # Add documents
                self.documents.add(list(documents))

        # Return unmodified input documents
        return documents

    def addobject(self, data, uid, field):
        """
        Helper method that builds a batch of object documents.

        Args:
            data: object content
            uid: optional list of corresponding uids
            field: optional field to set

        Returns:
            documents
        """

        # Raise error if index is not writable
        if not self.config.get("writable"):
            raise ReadOnlyError("Attempting to add documents to a read-only index (writable != True)")

        documents = []
        for x, content in enumerate(data):
            if field:
                row = {"id": uid[x], field: content} if uid else {field: content}
            elif uid:
                row = (uid[x], content)
            else:
                row = content

            documents.append(row)

        return self.add(documents)

    def index(self):
        """
        Builds an embeddings index for previously batched documents.
        """

        # Raise error if index is not writable
        if not self.config.get("writable"):
            raise ReadOnlyError("Attempting to index a read-only index (writable != True)")

        if self.embeddings and self.documents:
            with self.lock:
                # Reset index
                self.indexes(False)

                # Build scoring index if term weighting is enabled
                if self.embeddings.isweighted():
                    self.embeddings.score(self.documents)

                # Build embeddings index
                self.embeddings.index(self.documents)

                # Save index if path available, otherwise this is an memory-only index
                if self.config.get("path"):
                    self.embeddings.save(self.config["path"], self.config.get("cloud"))

                # Reset document stream
                self.documents.close()
                self.documents = None

    def upsert(self):
        """
        Runs an embeddings upsert operation for previously batched documents.
        """

        # Raise error if index is not writable
        if not self.config.get("writable"):
            raise ReadOnlyError("Attempting to upsert a read-only index (writable != True)")

        if self.embeddings and self.documents:
            with self.lock:
                # Run upsert
                self.embeddings.upsert(self.documents)

                # Save index if path available, otherwise this is an memory-only index
                if self.config.get("path"):
                    self.embeddings.save(self.config["path"], self.config.get("cloud"))

                # Reset document stream
                self.documents.close()
                self.documents = None

    def delete(self, ids):
        """
        Deletes from an embeddings index. Returns list of ids deleted.

        Args:
            ids: list of ids to delete

        Returns:
            ids deleted
        """

        # Raise error if index is not writable
        if not self.config.get("writable"):
            raise ReadOnlyError("Attempting to delete from a read-only index (writable != True)")

        if self.embeddings:
            with self.lock:
                # Run delete operation
                deleted = self.embeddings.delete(ids)

                # Save index if path available, otherwise this is an memory-only index
                if self.config.get("path"):
                    self.embeddings.save(self.config["path"], self.config.get("cloud"))

                # Return deleted ids
                return deleted

        return None

    def reindex(self, config, function=None):
        """
        Recreates embeddings index using config. This method only works if document content storage is enabled.

        Args:
            config: new config
            function: optional function to prepare content for indexing
        """

        # Raise error if index is not writable
        if not self.config.get("writable"):
            raise ReadOnlyError("Attempting to reindex a read-only index (writable != True)")

        if self.embeddings:
            with self.lock:
                # Resolve function, if necessary
                function = self.function(function) if function and isinstance(function, str) else function

                # Reindex
                self.embeddings.reindex(config, function)

                # Save index if path available, otherwise this is an memory-only index
                if self.config.get("path"):
                    self.embeddings.save(self.config["path"], self.config.get("cloud"))

    def count(self):
        """
        Total number of elements in this embeddings index.

        Returns:
            number of elements in embeddings index
        """

        if self.embeddings:
            return self.embeddings.count()

        return None

    def similarity(self, query, texts):
        """
        Computes the similarity between query and list of text. Returns a list of
        {id: value, score: value} sorted by highest score, where id is the index
        in texts.

        Args:
            query: query text
            texts: list of text

        Returns:
            list of {id: value, score: value}
        """

        # Use similarity instance if available otherwise fall back to embeddings model
        if "similarity" in self.pipelines:
            return [{"id": uid, "score": float(score)} for uid, score in self.pipelines["similarity"](query, texts)]
        if self.embeddings:
            return [{"id": uid, "score": float(score)} for uid, score in self.embeddings.similarity(query, texts)]

        return None

    def batchsimilarity(self, queries, texts):
        """
        Computes the similarity between list of queries and list of text. Returns a list
        of {id: value, score: value} sorted by highest score per query, where id is the
        index in texts.

        Args:
            queries: queries text
            texts: list of text

        Returns:
            list of {id: value, score: value} per query
        """

        # Use similarity instance if available otherwise fall back to embeddings model
        if "similarity" in self.pipelines:
            return [[{"id": uid, "score": float(score)} for uid, score in r] for r in self.pipelines["similarity"](queries, texts)]
        if self.embeddings:
            return [[{"id": uid, "score": float(score)} for uid, score in r] for r in self.embeddings.batchsimilarity(queries, texts)]

        return None

    def explain(self, query, texts=None, limit=10):
        """
        Explains the importance of each input token in text for a query.

        Args:
            query: query text
            texts: optional list of text, otherwise runs search query
            limit: optional limit if texts is None

        Returns:
            list of dict per input text where a higher token scores represents higher importance relative to the query
        """

        if self.embeddings:
            with self.lock:
                return self.embeddings.explain(query, texts, limit)

        return None

    def batchexplain(self, queries, texts=None, limit=10):
        """
        Explains the importance of each input token in text for a list of queries.

        Args:
            query: queries text
            texts: optional list of text, otherwise runs search queries
            limit: optional limit if texts is None

        Returns:
            list of dict per input text per query where a higher token scores represents higher importance relative to the query
        """

        if self.embeddings:
            with self.lock:
                return self.embeddings.batchexplain(queries, texts, limit)

        return None

    def transform(self, text, category=None, index=None):
        """
        Transforms text into embeddings arrays.

        Args:
            text: input text
            category: category for instruction-based embeddings
            index: index name, if applicable

        Returns:
            embeddings array
        """

        if self.embeddings:
            return [float(x) for x in self.embeddings.transform(text, category, index)]

        return None

    def batchtransform(self, texts, category=None, index=None):
        """
        Transforms list of text into embeddings arrays.

        Args:
            texts: list of text
            category: category for instruction-based embeddings
            index: index name, if applicable

        Returns:
            embeddings arrays
        """

        if self.embeddings:
            return [[float(x) for x in result] for result in self.embeddings.batchtransform(texts, category, index)]

        return None

    def extract(self, queue, texts=None):
        """
        Extracts answers to input questions.

        Args:
            queue: list of {name: value, query: value, question: value, snippet: value}
            texts: optional list of text

        Returns:
            list of {name: value, answer: value}
        """

        if self.embeddings and "extractor" in self.pipelines:
            # Get extractor instance
            extractor = self.pipelines["extractor"]

            # Run extractor and return results as dicts
            return extractor(queue, texts)

        return None

    def label(self, text, labels):
        """
        Applies a zero shot classifier to text using a list of labels. Returns a list of
        {id: value, score: value} sorted by highest score, where id is the index in labels.

        Args:
            text: text|list
            labels: list of labels

        Returns:
            list of {id: value, score: value} per text element
        """

        if "labels" in self.pipelines:
            # Text is a string
            if isinstance(text, str):
                return [{"id": uid, "score": float(score)} for uid, score in self.pipelines["labels"](text, labels)]

            # Text is a list
            return [[{"id": uid, "score": float(score)} for uid, score in result] for result in self.pipelines["labels"](text, labels)]

        return None

    def pipeline(self, name, *args, **kwargs):
        """
        Generic pipeline execution method.

        Args:
            name: pipeline name
            args: pipeline positional arguments
            kwargs: pipeline keyword arguments

        Returns:
            pipeline results
        """

        # Backwards compatible with previous pipeline function arguments
        args = args[0] if args and len(args) == 1 and isinstance(args[0], tuple) else args

        if name in self.pipelines:
            return self.pipelines[name](*args, **kwargs)

        return None

    def workflow(self, name, elements):
        """
        Executes a workflow.

        Args:
            name: workflow name
            elements: elements to process

        Returns:
            processed elements
        """

        if hasattr(elements, "__len__") and hasattr(elements, "__getitem__"):
            # Convert to tuples and return as a list since input is sized
            elements = [tuple(element) if isinstance(element, list) else element for element in elements]
        else:
            # Convert to tuples and return as a generator since input is not sized
            elements = (tuple(element) if isinstance(element, list) else element for element in elements)

        # Execute workflow
        return self.workflows[name](elements)

    def agent(self, name, *args, **kwargs):
        """
        Executes an agent.

        Args:
            name: agent name
            args: agent positional arguments
            kwargs: agent keyword arguments
        """

        if name in self.agents:
            return self.agents[name](*args, **kwargs)

        return None

    def wait(self):
        """
        Closes threadpool and waits for completion.
        """

        if self.pool:
            self.pool.close()
            self.pool.join()
            self.pool = None


class ReadOnlyError(Exception):
    """
    Error raised when trying to modify a read-only index
    """



"""
Faiss module
"""

import math
import platform

import numpy as np

from faiss import omp_set_num_threads
from faiss import index_factory, IO_FLAG_MMAP, METRIC_INNER_PRODUCT, read_index, write_index
from faiss import index_binary_factory, read_index_binary, write_index_binary, IndexBinaryIDMap

from ..base import ANN

if platform.system() == "Darwin":
    # Workaround for a Faiss issue causing segmentation faults on macOS. See txtai FAQ for more.
    omp_set_num_threads(1)


class Faiss(ANN):
    """
    Builds an ANN index using the Faiss library.
    """

    def __init__(self, config):
        super().__init__(config)

        # Scalar quantization
        quantize = self.config.get("quantize")
        self.qbits = quantize if quantize and isinstance(quantize, int) and not isinstance(quantize, bool) else None

    def load(self, path):
        # Get read function
        readindex = read_index_binary if self.qbits else read_index

        # Load index
        self.backend = readindex(path, IO_FLAG_MMAP if self.setting("mmap") is True else 0)

    def index(self, embeddings):
        # Compute model training size
        train, sample = embeddings, self.setting("sample")
        if sample:
            # Get sample for training
            rng = np.random.default_rng(0)
            indices = sorted(rng.choice(train.shape[0], int(sample * train.shape[0]), replace=False, shuffle=False))
            train = train[indices]

        # Configure embeddings index. Inner product is equal to cosine similarity on normalized vectors.
        params = self.configure(embeddings.shape[0], train.shape[0])

        # Create index
        self.backend = self.create(embeddings, params)

        # Train model
        self.backend.train(train)

        # Add embeddings - position in embeddings is used as the id
        self.backend.add_with_ids(embeddings, np.arange(embeddings.shape[0], dtype=np.int64))

        # Add id offset and index build metadata
        self.config["offset"] = embeddings.shape[0]
        self.metadata({"components": params})

    def append(self, embeddings):
        new = embeddings.shape[0]

        # Append new ids - position in embeddings + existing offset is used as the id
        self.backend.add_with_ids(embeddings, np.arange(self.config["offset"], self.config["offset"] + new, dtype=np.int64))

        # Update id offset and index metadata
        self.config["offset"] += new
        self.metadata()

    def delete(self, ids):
        # Remove specified ids
        self.backend.remove_ids(np.array(ids, dtype=np.int64))

    def search(self, queries, limit):
        # Set nprobe and nflip search parameters
        self.backend.nprobe = self.nprobe()
        self.backend.nflip = self.setting("nflip", self.backend.nprobe)

        # Run the query
        scores, ids = self.backend.search(queries, limit)

        # Map results to [(id, score)]
        results = []
        for x, score in enumerate(scores):
            # Transform scores and add results
            results.append(list(zip(ids[x].tolist(), self.scores(score))))

        return results

    def count(self):
        return self.backend.ntotal

    def save(self, path):
        # Get write function
        writeindex = write_index_binary if self.qbits else write_index

        # Write index
        writeindex(self.backend, path)

    def configure(self, count, train):
        """
        Configures settings for a new index.

        Args:
            count: initial number of embeddings rows
            train: number of rows selected for model training

        Returns:
            user-specified or generated components setting
        """

        # Lookup components setting
        components = self.setting("components")

        if components:
            # Format and return components string
            return self.components(components, train)

        # Derive quantization. Prefer backend-specific setting. Fallback to root-level parameter.
        quantize = self.setting("quantize", self.config.get("quantize"))
        quantize = 8 if isinstance(quantize, bool) else quantize

        # Get storage setting
        storage = f"SQ{quantize}" if quantize else "Flat"

        # Small index, use storage directly with IDMap
        if count <= 5000:
            return "BFlat" if self.qbits else f"IDMap,{storage}"

        x = self.cells(train)
        components = f"BIVF{x}" if self.qbits else f"IVF{x},{storage}"

        return components

    def create(self, embeddings, params):
        """
        Creates a new index.

        Args:
            embeddings: embeddings to index
            params: index parameters

        Returns:
            new index
        """

        # Create binary index
        if self.qbits:
            index = index_binary_factory(embeddings.shape[1] * 8, params)

            # Wrap with BinaryIDMap, if necessary
            if any(x in params for x in ["BFlat", "BHNSW"]):
                index = IndexBinaryIDMap(index)

            return index

        # Create standard float index
        return index_factory(embeddings.shape[1], params, METRIC_INNER_PRODUCT)

    def cells(self, count):
        """
        Calculates the number of IVF cells for an IVF index.

        Args:
            count: number of embeddings rows

        Returns:
            number of IVF cells
        """

        # Calculate number of IVF cells where x = min(4 * sqrt(embeddings count), embeddings count / 39)
        # Faiss requires at least 39 points per cluster
        return max(min(round(4 * math.sqrt(count)), int(count / 39)), 1)

    def components(self, components, train):
        """
        Formats a components string. This method automatically calculates the optimal number of IVF cells, if omitted.

        Args:
            components: input components string
            train: number of rows selected for model training

        Returns:
            formatted components string
        """

        # Optimal number of IVF cells
        x = self.cells(train)

        # Add number of IVF cells, if missing
        components = [f"IVF{x}" if component == "IVF" else component for component in components.split(",")]

        # Return components string
        return ",".join(components)

    def nprobe(self):
        """
        Gets or derives the nprobe search parameter.

        Returns:
            nprobe setting
        """

        # Get size of embeddings index
        count = self.count()

        default = 6 if count <= 5000 else round(self.cells(count) / 16)
        return self.setting("nprobe", default)

    def scores(self, scores):
        """
        Calculates the index score from the input score. This method returns the hamming score
        (1.0 - (hamming distance / total number of bits)) for binary indexes and the input
        scores otherwise.

        Args:
            scores: input scores

        Returns:
            index scores
        """

        # Calculate hamming score, bound between 0.0 - 1.0
        if self.qbits:
            return np.clip(1.0 - (scores / (self.config["dimensions"] * 8)), 0.0, 1.0).tolist()

        # Standard scoring
        return scores.tolist()



"""
Compress module
"""

import os


class Compress:
    """
    Base class for Compress instances.
    """

    def pack(self, path, output):
        """
        Compresses files in directory path to file output.

        Args:
            path: input directory path
            output: output file
        """

        raise NotImplementedError

    def unpack(self, path, output):
        """
        Extracts all files in path to output.

        Args:
            path: input file path
            output: output directory
        """

        raise NotImplementedError

    def validate(self, directory, path):
        """
        Validates path is under directory.

        Args:
            directory: base directory
            path: path to validate

        Returns:
            True if path is under directory, False otherwise
        """

        directory = os.path.abspath(directory)
        path = os.path.abspath(path)
        prefix = os.path.commonprefix([directory, path])

        return prefix == directory



"""
Stream module
"""

from .autoid import AutoId
from .transform import Action


class Stream:
    """
    Yields input document as standard (id, data, tags) tuples.
    """

    def __init__(self, embeddings, action=None):
        """
        Create a new stream.

        Args:
            embeddings: embeddings instance
            action: optional index action
        """

        self.embeddings = embeddings
        self.action = action

        # Alias embeddings attributes
        self.config = embeddings.config

        # Get config parameters
        self.offset = self.config.get("offset", 0) if action == Action.UPSERT else 0
        autoid = self.config.get("autoid", self.offset)

        # Create autoid generator, reset int sequence if this isn't an UPSERT
        autoid = 0 if isinstance(autoid, int) and action != Action.UPSERT else autoid
        self.autoid = AutoId(autoid)

    def __call__(self, documents):
        """
        Yield (id, data, tags) tuples from a stream of documents.

        Args:
            documents: input documents
        """

        # Iterate over documents and yield standard (id, data, tag) tuples
        for document in documents:
            if isinstance(document, dict):
                # Create (id, data, tags) tuple from dictionary
                document = document.get("id"), document, document.get("tags")
            elif isinstance(document, tuple):
                # Create (id, data, tags) tuple
                document = document if len(document) >= 3 else (document[0], document[1], None)
            else:
                # Create (id, data, tags) tuple with empty fields
                document = None, document, None

            # Set autoid if the action is set
            if self.action and document[0] is None:
                document = (self.autoid(document[1]), document[1], document[2])

            # Yield (id, data, tags) tuple
            yield document

        # Save autoid sequence if used
        current = self.autoid.current()
        if self.action and current:
            self.config["autoid"] = current



"""
Placeholder module
"""


class Agent:
    """
    Agent placeholder stub for when smolagents isn't installed
    """

    def __init__(self, *args, **kwargs):
        """
        Raises an exception that smolagents isn't installed.
        """

        raise ImportError('smolagents is not available - install "agent" extra to enable')



"""
Extension module
"""


class Extension:
    """
    Defines an API extension. API extensions can expose custom pipelines or other custom logic.
    """

    def __call__(self, app):
        """
        Hook to register custom routing logic and/or modify the FastAPI instance.

        Args:
            app: FastAPI application instance
        """

        return



"""
Configuration module
"""

import json
import os

from ...serialize import SerializeFactory


class Configuration:
    """
    Loads and saves index configuration.
    """

    def load(self, path):
        """
        Loads index configuration. This method supports both config.json and config pickle files.

        Args:
            path: path to directory

        Returns:
            dict
        """

        # Configuration
        config = None

        # Determine if config is json or pickle
        jsonconfig = os.path.exists(f"{path}/config.json")

        # Set config file name
        name = "config.json" if jsonconfig else "config"

        # Load configuration
        with open(f"{path}/{name}", "r" if jsonconfig else "rb", encoding="utf-8" if jsonconfig else None) as handle:
            # Load JSON, also backwards-compatible with pickle configuration
            config = json.load(handle) if jsonconfig else SerializeFactory.create("pickle").loadstream(handle)

        # Add format parameter
        config["format"] = "json" if jsonconfig else "pickle"

        return config

    def save(self, config, path):
        """
        Saves index configuration. This method defaults to JSON and falls back to pickle.

        Args:
            config: configuration to save
            path: path to directory

        Returns:
            dict
        """

        # Default to JSON config
        jsonconfig = config.get("format", "json") == "json"

        # Set config file name
        name = "config.json" if jsonconfig else "config"

        # Write configuration
        with open(f"{path}/{name}", "w" if jsonconfig else "wb", encoding="utf-8" if jsonconfig else None) as handle:
            if jsonconfig:
                # Write config as JSON
                json.dump(config, handle, default=str, indent=2)
            else:
                # Backwards compatible method to save pickle configuration
                SerializeFactory.create("pickle").savestream(config, handle)



"""
Recovery module
"""

import os
import shutil


class Recovery:
    """
    Vector embeddings recovery. This class handles streaming embeddings from a vector checkpoint file.
    """

    def __init__(self, checkpoint, vectorsid, load):
        """
        Creates a Recovery instance.

        Args:
            checkpoint: checkpoint directory
            vectorsid: vectors uid for current configuration
            load: load embeddings method
        """

        self.spool, self.path, self.load = None, None, load

        # Get unique file id
        path = f"{checkpoint}/{vectorsid}"
        if os.path.exists(path):
            # Generate recovery path
            self.path = f"{checkpoint}/recovery"

            # Copy current checkpoint to recovery
            shutil.copyfile(path, self.path)

            # Open file an return
            # pylint: disable=R1732
            self.spool = open(self.path, "rb")

    def __call__(self):
        """
        Reads and returns the next batch of embeddings.

        Returns
            batch of embeddings
        """

        try:
            return self.load(self.spool) if self.spool else None
        except EOFError:
            # End of spool file, cleanup
            self.spool.close()
            os.remove(self.path)

            # Clear parameters
            self.spool, self.path = None, None

            return None



"""
Sentence Transformers module
"""

# Conditional import
try:
    from sentence_transformers import SentenceTransformer

    SENTENCE_TRANSFORMERS = True
except ImportError:
    SENTENCE_TRANSFORMERS = False

from ...models import Models

from ..base import Vectors


class STVectors(Vectors):
    """
    Builds vectors using sentence-transformers (aka SBERT).
    """

    def __init__(self, config, scoring, models):
        # Check before parent constructor since it calls loadmodel
        if not SENTENCE_TRANSFORMERS:
            raise ImportError('sentence-transformers is not available - install "vectors" extra to enable')

        # Pool parameter created here since loadmodel is called from parent constructor
        self.pool = None

        super().__init__(config, scoring, models)

    def loadmodel(self, path):
        # Get target device
        gpu, pool = self.config.get("gpu", True), False

        # Default mode uses a single GPU. Setting to all spawns a process per GPU.
        if isinstance(gpu, str) and gpu == "all":
            # Get number of accelerator devices available
            devices = Models.acceleratorcount()

            # Enable multiprocessing pooling only when multiple devices are available
            gpu, pool = devices <= 1, devices > 1

        # Tensor device id
        deviceid = Models.deviceid(gpu)

        # Additional model arguments
        modelargs = self.config.get("vectors", {})

        # Load sentence-transformers encoder
        model = self.loadencoder(path, device=Models.device(deviceid), **modelargs)

        # Start process pool for multiple GPUs
        if pool:
            self.pool = model.start_multi_process_pool()

        # Return model
        return model

    def encode(self, data, category=None):
        # Get encode method based on input category
        encode = self.model.encode_query if category == "query" else self.model.encode_document if category == "data" else self.model.encode

        # Additional encoding arguments
        encodeargs = self.config.get("encodeargs", {})

        # Encode with sentence transformers encoder
        return encode(data, pool=self.pool, batch_size=self.encodebatch, **encodeargs)

    def close(self):
        # Close pool before model is closed in parent method
        if self.pool:
            self.model.stop_multi_process_pool(self.pool)
            self.pool = None

        super().close()

    def loadencoder(self, path, device, **kwargs):
        """
        Loads the embeddings encoder model from path.

        Args:
            path: model path
            device: tensor device
            kwargs: additional keyword args

        Returns:
            embeddings encoder
        """

        return SentenceTransformer(path, device=device, **kwargs)



"""
Embedded module
"""

from .rdbms import RDBMS


class Embedded(RDBMS):
    """
    Base class for embedded relational databases. An embedded relational database stores all content in a local file.
    """

    def __init__(self, config):
        """
        Creates a new Database.

        Args:
            config: database configuration parameters
        """

        super().__init__(config)

        # Path to database file
        self.path = None

    def load(self, path):
        # Call parent logic
        super().load(path)

        # Store path reference
        self.path = path

    def save(self, path):
        # Temporary database
        if not self.path:
            # Save temporary database
            self.connection.commit()

            # Copy data from current to new
            connection = self.copy(path)

            # Close temporary database
            self.connection.close()

            # Point connection to new connection
            self.session(connection=connection)
            self.path = path

        # Paths are equal, commit changes
        elif self.path == path:
            self.connection.commit()

        # New path is different from current path, copy data and continue using current connection
        else:
            self.copy(path).close()

    def jsonprefix(self):
        # Return json column prefix
        return "json_extract(data"

    def jsoncolumn(self, name):
        # Generate json column using json_extract function
        return f"json_extract(data, '$.{name}')"

    def copy(self, path):
        """
        Copies the current database into path.

        Args:
            path: path to write database

        Returns:
            new connection with data copied over
        """

        raise NotImplementedError



"""
Muvera module
"""

import numpy as np


class Muvera:
    """
    Implements the MUVERA (Multi-Vector Retrieval via Fixed Dimensional Encodings) algorithm. This reduces late interaction multi-vector
    outputs to a single fixed vector.

    The size of the output vectors are set using the following parameters

        output dimensions = repetitions * 2^hashes * projected

    For example, the default parameters create vectors with the following output dimensions.

        output dimensions = 20 * 2^5 * 16 = 10240

    This code is based on the following:
      - Paper: https://arxiv.org/abs/2405.19504
      - GitHub: https://github.com/google/graph-mining/tree/main/sketching/point_cloud
      - Python port of the original C++ code: https://github.com/sigridjineth/muvera-py
    """

    def __init__(self, repetitions=20, hashes=5, projection=16, seed=42):
        """
        Creates a Muvera instance.

        Args:
            repetitions: number of iterations
            hashes: number of simhash partitions as 2^hashes
            projection: dimensionality reduction, uses an identity projection when set to None
            seed: random seed
        """

        # Number of repetitions
        self.repetitions = repetitions

        # Number of simhash projections
        self.hashes = hashes

        # Optional number of projected dimensions
        self.projection = projection

        # Seed
        self.seed = seed

    def __call__(self, data, category):
        """
        Transforms a list of multi-vector collections into single fixed vector outputs.

        Args:
            data: array of multi-vector vectors
            category: embeddings category (query or data)
        """

        # Get stats
        dimension, length = data[0].shape[1], len(data)

        # Determine projection dimension
        identity = not self.projection
        projection = dimension if identity else self.projection

        # Number of simhash partitions
        partitions = 2**self.hashes

        # Document tracking
        lengths = np.array([len(doc) for doc in data], dtype=np.int32)
        total = np.sum(lengths)
        documents = np.repeat(np.arange(length), lengths)

        # Stack all vectors
        points = np.vstack(data).astype(np.float32)

        # Output vectors
        size = self.repetitions * partitions * projection
        vectors = np.zeros((length, size), dtype=np.float32)

        # Process each repetition
        for number in range(self.repetitions):
            seed = self.seed + number

            # Calculate the simhash
            sketches = points @ self.random(dimension, self.hashes, seed)

            # Dimensionality reduction, if necessary
            projected = points if identity else (points @ self.reducer(dimension, projection, seed))

            # Get partition indices
            bits = (sketches > 0).astype(np.uint32)
            indices = np.zeros(total, dtype=np.uint32)

            # Calculate vector indices
            for x in range(self.hashes):
                indices = (indices << 1) + (bits[:, x] ^ (indices & 1))

            # Initialize storage
            fdesum = np.zeros((length * partitions * projection,), dtype=np.float32)
            counts = np.zeros((length, partitions), dtype=np.int32)

            # Count vectors per partition per document
            np.add.at(counts, (documents, indices), 1)

            # Aggregate vectors using flattened indexing for efficiency
            part = documents * partitions + indices
            base = part * projection

            for d in range(projection):
                flat = base + d
                np.add.at(fdesum, flat, projected[:, d])

            # Reshape for easier manipulation
            # pylint: disable=E1121
            fdesum = fdesum.reshape(length, partitions, projection)

            # Convert sums to averages for data category
            if category == "data":
                # Safe division (avoid divide by zero)
                counts = counts[:, :, np.newaxis]
                np.divide(fdesum, counts, out=fdesum, where=counts > 0)

            # Save results
            start = number * partitions * projection
            vectors[:, start : start + partitions * projection] = fdesum.reshape(length, -1)

        return vectors

    def random(self, dimension, projection, seed):
        """
        Generates a random matrix for simhash projections.

        Args:
            dimensions: number of dimensions for input vectors
            projections: number of projection dimensions
            seed: random seed

        Returns:
            random matrix for simhash projections
        """

        rng = np.random.default_rng(seed)
        return rng.normal(loc=0.0, scale=1.0, size=(dimension, projection)).astype(np.float32)

    def reducer(self, dimension, projection, seed):
        """
        Generates a random matrix for dimensionality reduction using the AMS sketch algorithm.

        Args:
            dimension: number of input dimensions
            projected: number of dimensions to project inputs to

        Returns:
            Dimensionality reduced matrix
        """

        rng = np.random.default_rng(seed)
        out = np.zeros((dimension, projection), dtype=np.float32)
        indices = rng.integers(0, projection, size=dimension)
        signs = rng.choice([-1.0, 1.0], size=dimension)
        out[np.arange(dimension), indices] = signs

        return out



"""
Annoy module
"""

# Conditional import
try:
    from annoy import AnnoyIndex

    ANNOY = True
except ImportError:
    ANNOY = False

from ..base import ANN


# pylint: disable=W0223
class Annoy(ANN):
    """
    Builds an ANN index using the Annoy library.
    """

    def __init__(self, config):
        super().__init__(config)

        if not ANNOY:
            raise ImportError('Annoy is not available - install "ann" extra to enable')

    def load(self, path):
        # Load index
        self.backend = AnnoyIndex(self.config["dimensions"], self.config["metric"])
        self.backend.load(path)

    def index(self, embeddings):
        # Inner product is equal to cosine similarity on normalized vectors
        self.config["metric"] = "dot"

        # Create index
        self.backend = AnnoyIndex(self.config["dimensions"], self.config["metric"])

        # Add items - position in embeddings is used as the id
        for x in range(embeddings.shape[0]):
            self.backend.add_item(x, embeddings[x])

        # Build index
        ntrees = self.setting("ntrees", 10)
        self.backend.build(ntrees)

        # Add index build metadata
        self.metadata({"ntrees": ntrees})

    def search(self, queries, limit):
        # Lookup search k setting
        searchk = self.setting("searchk", -1)

        # Annoy doesn't have a built in batch query method
        results = []
        for query in queries:
            # Run the query
            ids, scores = self.backend.get_nns_by_vector(query, n=limit, search_k=searchk, include_distances=True)

            # Map results to [(id, score)]
            results.append(list(zip(ids, scores)))

        return results

    def count(self):
        # Number of items in index
        return self.backend.get_n_items()

    def save(self, path):
        # Write index
        self.backend.save(path)



"""
Factory module
"""

from ...util import Resolver

from .ivfsparse import IVFSparse
from .pgsparse import PGSparse


class SparseANNFactory:
    """
    Methods to create Sparse ANN indexes.
    """

    @staticmethod
    def create(config):
        """
        Create an Sparse ANN.

        Args:
            config: index configuration parameters

        Returns:
            Sparse ANN
        """

        # ANN instance
        ann = None
        backend = config.get("backend", "ivfsparse")

        # Create ANN instance
        if backend == "ivfsparse":
            ann = IVFSparse(config)
        elif backend == "pgsparse":
            ann = PGSparse(config)
        else:
            ann = SparseANNFactory.resolve(backend, config)

        # Store config back
        config["backend"] = backend

        return ann

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: index configuration parameters

        Returns:
            ANN
        """

        try:
            return Resolver()(backend)(config)
        except Exception as e:
            raise ImportError(f"Unable to resolve sparse ann backend: '{backend}'") from e



"""
Generation module
"""

import re

from ...util import TemplateFormatter


class Generation:
    """
    Base class for generative models. This class has common logic for building prompts and cleaning model results.
    """

    def __init__(self, path=None, template=None, **kwargs):
        """
        Creates a new Generation instance.

        Args:
            path: model path
            template: prompt template
            kwargs: additional keyword arguments
        """

        self.path = path
        self.template = template
        self.kwargs = kwargs

    def __call__(self, text, maxlength, stream, stop, defaultrole, stripthink, **kwargs):
        """
        Generates text. Supports the following input formats:

          - String or list of strings (instruction-tuned models must follow chat templates)
          - List of dictionaries with `role` and `content` key-values or lists of lists

        Args:
            text: text|list
            maxlength: maximum sequence length
            stream: stream response if True, defaults to False
            stop: list of stop strings
            defaultrole: default role to apply to text inputs (prompt for raw prompts (default) or user for user chat messages)
            stripthink: strip thinking tags, defaults to False
            kwargs: additional generation keyword arguments

        Returns:
            generated text
        """

        # Format inputs
        texts = [text] if isinstance(text, str) or isinstance(text[0], dict) else text

        # Apply template, if necessary
        if self.template:
            formatter = TemplateFormatter()
            texts = [formatter.format(self.template, text=x) if isinstance(x, str) else x for x in texts]

        # Apply default role, if necessary
        if defaultrole == "user":
            texts = [[{"role": "user", "content": x}] if isinstance(x, str) else x for x in texts]

        # Run pipeline
        results = self.execute(texts, maxlength, stream, stop, **kwargs)

        # Streaming generation
        if stream:
            return results

        # Clean generated text
        results = [self.clean(texts[x], result, stripthink) for x, result in enumerate(results)]

        # Extract results based on inputs
        return results[0] if isinstance(text, str) or isinstance(text[0], dict) else results

    def isvision(self):
        """
        Returns True if this LLM supports vision operations.

        Returns:
            True if this is a vision model
        """

        return False

    def execute(self, texts, maxlength, stream, stop, **kwargs):
        """
        Runs a list of prompts through a generative model.

        Args:
            texts: list of prompts to run
            maxlength: maximum sequence length
            stream: stream response if True, defaults to False
            stop: list of stop strings
            kwargs: additional generation keyword arguments

        Returns:
            generated text
        """

        # Streaming generation
        if stream:
            return self.stream(texts, maxlength, stream, stop, **kwargs)

        # Full response as content elements
        return list(self.stream(texts, maxlength, stream, stop, **kwargs))

    def clean(self, prompt, result, stripthink):
        """
        Applies a series of rules to clean generated text.

        Args:
            prompt: original input prompt
            result: result text
            stripthink: removes thinking tags if true

        Returns:
            clean text
        """

        # Replace input prompt
        text = result.replace(prompt, "") if isinstance(prompt, str) else result

        # Replace thinking tags, if necessary
        text = re.sub(r"(?s)<think>.+?</think>", "", text).strip() if stripthink else text

        # Apply text cleaning rules
        return text.replace("$=", "<=").strip()

    def response(self, result):
        """
        Parses response content from the result. This supports both standard and streaming
        generation.

        For standard generation, the full response is returned. For streaming generation,
        this method will stream chunks of content.

        Args:
            result: LLM response

        Returns:
            response
        """

        streamed = False
        for chunk in result:
            # Expects one of the following parameter paths
            #  - text
            #  - message.content
            #  - delta.content
            data = chunk["choices"][0]
            text = data.get("text", data.get("message", data.get("delta")))
            text = text if isinstance(text, str) else text.get("content")

            # Yield result if there is text AND it's not leading stream whitespace
            if text is not None and (streamed or text.strip()):
                yield (text.lstrip() if not streamed else text)
                streamed = True

    def stream(self, texts, maxlength, stream, stop, **kwargs):
        """
        Streams LLM responses.

        Args:
            texts: list of prompts to run
            maxlength: maximum sequence length
            stream: stream response if True, defaults to False
            stop: list of stop strings
            kwargs: additional generation keyword arguments

        Returns:
            responses
        """

        raise NotImplementedError



"""
Serialize module
"""


class Serialize:
    """
    Base class for Serialize instances. This class serializes data to files, streams and bytes.
    """

    def load(self, path):
        """
        Loads data from path.

        Args:
            path: input path

        Returns:
            deserialized data
        """

        with open(path, "rb") as handle:
            return self.loadstream(handle)

    def save(self, data, path):
        """
        Saves data to path.

        Args:
            data: data to save
            path: output path
        """

        with open(path, "wb") as handle:
            self.savestream(data, handle)

    def loadstream(self, stream):
        """
        Loads data from stream.

        Args:
            stream: input stream

        Returns:
            deserialized data
        """

        raise NotImplementedError

    def savestream(self, data, stream):
        """
        Saves data to stream.

        Args:
            data: data to save
            stream: output stream
        """

        raise NotImplementedError

    def loadbytes(self, data):
        """
        Loads data from bytes.

        Args:
            data: input bytes

        Returns:
            deserialized data
        """

        raise NotImplementedError

    def savebytes(self, data):
        """
        Saves data as bytes.

        Args:
            data: data to save

        Returns:
            serialized data
        """

        raise NotImplementedError



"""
Template module
"""

from string import Formatter


class TemplateFormatter(Formatter):
    """
    Custom Formatter that requires each argument to be consumed.
    """

    def check_unused_args(self, used_args, args, kwargs):
        difference = set(kwargs).difference(used_args)
        if difference:
            raise KeyError(difference)



"""
Signal module
"""

import numpy as np

# Conditional import
try:
    from scipy import signal
    from scipy.fft import rfft, rfftfreq

    SCIPY = True
except ImportError:
    SCIPY = False


class Signal:
    """
    Utility methods for audio signal processing.
    """

    @staticmethod
    def mono(audio):
        """
        Convert stereo to mono audio.

        Args:
            audio: audio data

        Returns:
            audio data with a single channel
        """

        return audio.mean(axis=1) if len(audio.shape) > 1 else audio

    @staticmethod
    def resample(audio, rate, target):
        """
        Resample audio if the sample rate doesn't match the target sample rate.

        Args:
            audio: audio data
            rate: current sample rate
            target: target sample rate

        Returns:
            audio resampled if necessary or original audio
        """

        if rate != target:
            # Transpose audio
            audio = audio.T

            # Resample audio and tranpose back
            samples = round(len(audio) * float(target) / rate)
            audio = signal.resample(audio, samples).T

        return audio

    @staticmethod
    def float32(audio):
        """
        Converts an input NumPy array with 16-bit ints to 32-bit floats.

        Args:
            audio: input audio array as 16-bit ints

        Returns:
            audio array as 32-bit floats
        """

        i = np.iinfo(audio.dtype)
        abs_max = 2 ** (i.bits - 1)
        offset = i.min + abs_max
        return (audio.astype(np.float32) - offset) / abs_max

    @staticmethod
    def int16(audio):
        """
        Converts an input NumPy array with 32-bit floats to 16-bit ints.

        Args:
            audio: input audio array as 32-bit floats

        Returns:
            audio array as 16-bit ints
        """

        i = np.iinfo(np.int16)
        absmax = 2 ** (i.bits - 1)
        offset = i.min + absmax
        return (audio * absmax + offset).clip(i.min, i.max).astype(np.int16)

    @staticmethod
    def mix(audio1, audio2, scale1=1, scale2=1):
        """
        Mixes audio1 and audio 2 into a single output audio segment.

        Args:
            audio1: audio segment 1
            audio2: audio segment 2
            scale1: scale factor for audio segment 1
            scale2: scale factor for audio segment 2
        """

        # Reshape audio, as necessary
        audio1 = audio1.reshape(1, -1) if len(audio1.shape) <= 1 else audio1
        audio2 = audio2.reshape(1, -1) if len(audio2.shape) <= 1 else audio2

        # Scale audio
        audio1 = audio1 * scale1
        audio2 = audio2 * scale2

        # Make audio files the same length
        large, small = (audio1, audio2) if audio1.shape[1] > audio2.shape[1] else (audio2, audio1)
        small = np.tile(small, (large.shape[1] // small.shape[1]) + 1).take(axis=1, indices=range(0, large.shape[1]))

        # Mix audio together
        return small + large

    @staticmethod
    def energy(audio, rate):
        """
        Calculates the signal energy for the input audio. Energy is defined as:

          Energy = 2 * Signal Amplitude

        Args:
            audio: audio data
            rate: sample rate

        Returns:
            {frequency: energy at that frequency}
        """

        # Calculate signal frequency
        frequency = rfftfreq(len(audio), 1.0 / rate)
        frequency = frequency[1:]

        # Calculate signal energy using amplitude
        energy = np.abs(rfft(audio))
        energy = energy[1:]
        energy = energy**2

        # Get energy for each frequency
        energyfreq = {}
        for x, freq in enumerate(frequency):
            if abs(freq) not in energyfreq:
                energyfreq[abs(freq)] = energy[x] * 2

        return energyfreq

    @staticmethod
    def trim(audio, rate, threshold=1, leading=True, trailing=True):
        """
        Removes leading and trailing silence from audio data.

        Args:
            audio: audio data
            rate: sample rate
            threshold: energy below this level will be considered silence, defaults to 1.0
            leading: trim leading silence, defaults to True
            trailing: trim trailing silence, defauls to True

        Returns:
            audio with silence removed
        """

        # Process in 20ms chunks
        n, offset = int(rate * (20 / 1000.0) * 2), 0

        chunks = []
        while offset + n <= len(audio):
            # Calculate energy for chunk and detection result
            chunk = audio[offset : offset + n]
            energyfreq = Signal.energy(chunk, rate)
            chunks.append((chunk, sum(energyfreq.values()) >= threshold))

            offset += n

        # Find first and last active chunks
        start = next((i for i, (_, active) in enumerate(chunks) if active), 0) if leading else 0
        end = (len(chunks) - next((i for i, (_, active) in enumerate(chunks[::-1]) if active), 0)) if trailing else len(chunks)

        # Concatenate active audio
        return np.concatenate([chunk for chunk, _ in chunks[start:end]])



"""
PGText module
"""

import os

# Conditional import
try:
    from sqlalchemy import create_engine, desc, delete, func, text
    from sqlalchemy import Column, Computed, Index, Integer, MetaData, StaticPool, Table, Text
    from sqlalchemy.dialects.postgresql import TSVECTOR
    from sqlalchemy.orm import Session
    from sqlalchemy.schema import CreateSchema

    PGTEXT = True
except ImportError:
    PGTEXT = False

from .base import Scoring


class PGText(Scoring):
    """
    Postgres full text search (FTS) based scoring.
    """

    def __init__(self, config=None):
        super().__init__(config)

        if not PGTEXT:
            raise ImportError('PGText is not available - install "scoring" extra to enable')

        # Database connection
        self.engine, self.database, self.connection, self.table = None, None, None, None

        # Language
        self.language = self.config.get("language", "english")

    def insert(self, documents, index=None, checkpoint=None):
        # Initialize tables
        self.initialize(recreate=True)

        # Collection of rows to insert
        rows = []

        # Collect rows
        for uid, document, _ in documents:
            # Extract text, if necessary
            if isinstance(document, dict):
                document = document.get(self.text, document.get(self.object))

            if document is not None:
                # If index is passed, use indexid, otherwise use id
                uid = index if index is not None else uid

                # Add row if the data type is accepted
                if isinstance(document, (str, list)):
                    rows.append((uid, " ".join(document) if isinstance(document, list) else document))

                # Increment index
                index = index + 1 if index is not None else None

        # Insert rows
        self.database.execute(self.table.insert(), [{"indexid": x, "text": text} for x, text in rows])

    def delete(self, ids):
        self.database.execute(delete(self.table).where(self.table.c["indexid"].in_(ids)))

    def weights(self, tokens):
        # Not supported
        return None

    def search(self, query, limit=3):
        # Run query
        query = (
            self.database.query(self.table.c["indexid"], text("ts_rank(vector, plainto_tsquery(:language, :query)) rank"))
            .order_by(desc(text("rank")))
            .limit(limit)
            .params({"language": self.language, "query": query})
        )

        return [(uid, score) for uid, score in query if score > 1e-5]

    def batchsearch(self, queries, limit=3, threads=True):
        return [self.search(query, limit) for query in queries]

    def count(self):
        # pylint: disable=E1102
        return self.database.query(func.count(self.table.c["indexid"])).scalar()

    def load(self, path):
        # Reset database to original checkpoint
        if self.database:
            self.database.rollback()
            self.connection.rollback()

        # Initialize tables
        self.initialize()

    def save(self, path):
        # Commit session and connection
        if self.database:
            self.database.commit()
            self.connection.commit()

    def close(self):
        if self.database:
            self.database.close()
            self.engine.dispose()

    def issparse(self):
        return True

    def isnormalized(self):
        return True

    def initialize(self, recreate=False):
        """
        Initializes a new database session.

        Args:
            recreate: Recreates the database tables if True
        """

        if not self.database:
            # Create engine, connection and session
            self.engine = create_engine(self.config.get("url", os.environ.get("SCORING_URL")), poolclass=StaticPool, echo=False)
            self.connection = self.engine.connect()
            self.database = Session(self.connection)

            # Set default schema, if necessary
            schema = self.config.get("schema")
            if schema:
                with self.engine.begin():
                    self.sqldialect(CreateSchema(schema, if_not_exists=True))

                self.sqldialect(text("SET search_path TO :schema"), {"schema": schema})

            # Table name
            table = self.config.get("table", "scoring")

            # Create vectors table
            self.table = Table(
                table,
                MetaData(),
                Column("indexid", Integer, primary_key=True, autoincrement=False),
                Column("text", Text),
                (
                    Column("vector", TSVECTOR, Computed(f"to_tsvector('{self.language}', text)", persisted=True))
                    if self.engine.dialect.name == "postgresql"
                    else Column("vector", Integer)
                ),
            )

            # Create text index
            index = Index(
                f"{table}-index",
                self.table.c["vector"],
                postgresql_using="gin",
            )

            # Drop and recreate table
            if recreate:
                self.table.drop(self.connection, checkfirst=True)
                index.drop(self.connection, checkfirst=True)

            # Create table and index
            self.table.create(self.connection, checkfirst=True)
            index.create(self.connection, checkfirst=True)

    def sqldialect(self, sql, parameters=None):
        """
        Executes a SQL statement based on the current SQL dialect.

        Args:
            sql: SQL to execute
            parameters: optional bind parameters
        """

        args = (sql, parameters) if self.engine.dialect.name == "postgresql" else (text("SELECT 1"),)
        self.database.execute(*args)



"""
Factory module
"""

from smolagents import CodeAgent, ToolCallingAgent

from .model import PipelineModel
from .tool import ToolFactory


class ProcessFactory:
    """
    Methods to create agent processes.
    """

    @staticmethod
    def create(config):
        """
        Create an agent process runner. The agent process runner takes a list of tools and an LLM
        and executes an agent process flow.

        Args:
            config: agent configuration

        Returns:
            agent process runner
        """

        constructor = ToolCallingAgent
        method = config.pop("method", None)
        if method == "code":
            constructor = CodeAgent

        # Create model backed by LLM pipeline
        model = config.pop("model", config.pop("llm", None))
        model = PipelineModel(**model) if isinstance(model, dict) else PipelineModel(model)

        # Create the agent process
        return constructor(tools=ToolFactory.create(config), model=model, **config)



"""
Defines API paths for llm endpoints.
"""

from typing import List, Optional

from fastapi import APIRouter, Body
from fastapi.responses import StreamingResponse

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/llm")
def llm(text: str, maxlength: Optional[int] = None, stream: Optional[bool] = False):
    """
    Runs a LLM pipeline for the input text.

    Args:
        text: input text
        maxlength: optional response max length
        stream: streams response if True

    Returns:
        response text
    """

    # Build keyword arguments
    kwargs = {key: value for key, value in [("stream", stream), ("maxlength", maxlength)] if value}

    # Run pipeline
    result = application.get().pipeline("llm", text, **kwargs)

    # Handle both standard and streaming responses
    return StreamingResponse(result) if stream else result


@router.post("/batchllm")
def batchllm(texts: List[str] = Body(...), maxlength: Optional[int] = Body(default=None), stream: Optional[bool] = Body(default=False)):
    """
    Runs a LLM pipeline for the input texts.

    Args:
        texts: input texts
        maxlength: optional response max length
        stream: streams response if True

    Returns:
        response texts
    """

    # Build keyword arguments
    kwargs = {key: value for key, value in [("stream", stream), ("maxlength", maxlength)] if value}

    # Run pipeline
    result = application.get().pipeline("llm", texts, **kwargs)

    # Handle both standard and streaming responses
    return StreamingResponse(result) if stream else result



"""
SIF module
"""

import numpy as np

from .tfidf import TFIDF


class SIF(TFIDF):
    """
    Smooth Inverse Frequency (SIF) scoring.
    """

    def __init__(self, config=None):
        super().__init__(config)

        # SIF configurable parameters
        self.a = self.config.get("a", 1e-3)

    def computefreq(self, tokens):
        # Default method computes frequency for a single entry
        # SIF uses word frequencies across entire index
        return {token: self.wordfreq[token] for token in tokens}

    def score(self, freq, idf, length):
        # Set freq to word frequencies across entire index when freq and idf shape don't match
        if isinstance(freq, np.ndarray) and freq.shape != np.array(idf).shape:
            freq.fill(freq.sum())

        # Calculate SIF score
        return self.a / (self.a + freq / self.tokens)



"""
Defines API paths for segmentation endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/segment")
def segment(text: str):
    """
    Segments text into semantic units.

    Args:
        text: input text

    Returns:
        segmented text
    """

    return application.get().pipeline("segmentation", (text,))


@router.post("/batchsegment")
def batchsegment(texts: List[str] = Body(...)):
    """
    Segments text into semantic units.

    Args:
        texts: list of texts to segment

    Returns:
        list of segmented text
    """

    return application.get().pipeline("segmentation", (texts,))



"""
Database module
"""

import logging
import types

from .encoder import EncoderFactory
from .sql import SQL, SQLError, Token

# Logging configuration
logger = logging.getLogger(__name__)


class Database:
    """
    Base class for database instances. This class encapsulates a content database used for
    storing field content as dicts and objects. The database instance works in conjuction
    with a vector index to execute SQL-driven similarity search.
    """

    def __init__(self, config):
        """
        Creates a new Database.

        Args:
            config: database configuration
        """

        # Initialize configuration
        self.configure(config)

    def load(self, path):
        """
        Loads a database path.

        Args:
            path: database url
        """

        raise NotImplementedError

    def insert(self, documents, index=0):
        """
        Inserts documents into the database.

        Args:
            documents: list of documents to save
            index: indexid offset, used for internal ids
        """

        raise NotImplementedError

    def delete(self, ids):
        """
        Deletes documents from database.

        Args:
            ids: ids to delete
        """

        raise NotImplementedError

    def reindex(self, config):
        """
        Reindexes internal database content and streams results back. This method must renumber indexids
        sequentially as deletes could have caused indexid gaps.

        Args:
            config: new configuration
        """

        raise NotImplementedError

    def save(self, path):
        """
        Saves a database at path.

        Args:
            path: path to write database
        """

        raise NotImplementedError

    def close(self):
        """
        Closes this database.
        """

        raise NotImplementedError

    def ids(self, ids):
        """
        Retrieves the internal indexids for a list of ids. Multiple indexids may be present for an id in cases
        where data is segmented.

        Args:
            ids: list of document ids

        Returns:
            list of (indexid, id)
        """

        raise NotImplementedError

    def count(self):
        """
        Retrieves the count of this database instance.

        Returns:
            total database count
        """

        raise NotImplementedError

    def search(self, query, similarity=None, limit=None, parameters=None, indexids=False):
        """
        Runs a search against the database. Supports the following methods:

            1. Standard similarity query. This mode retrieves content for the ids in the similarity results
            2. Similarity query as SQL. This mode will combine similarity results and database results into
               a single result set. Similarity queries are set via the SIMILAR() function.
            3. SQL with no similarity query. This mode runs a SQL query and retrieves the results without similarity queries.

        Example queries:
            "natural language processing" - standard similarity only query
            "select * from txtai where similar('natural language processing')" - similarity query as SQL
            "select * from txtai where similar('nlp') and entry > '2021-01-01'" - similarity query with additional SQL clauses
            "select id, text, score from txtai where similar('nlp')" - similarity query with additional SQL column selections
            "select * from txtai where entry > '2021-01-01' - database only query

        Args:
            query: input query
            similarity: similarity results as [(indexid, score)]
            limit: maximum number of results to return
            parameters: dict of named parameters to bind to placeholders

        Returns:
            query results as a list of dicts
            list of ([indexid, score]) if indexids is True
        """

        # Parse query if necessary
        if isinstance(query, str):
            query = self.parse(query)

        # Add in similar results
        where = query.get("where")

        if "select" in query and similarity:
            for x in range(len(similarity)):
                token = f"{Token.SIMILAR_TOKEN}{x}"
                if where and token in where:
                    where = where.replace(token, self.embed(similarity, x))

        elif similarity:
            # Not a SQL query, load similarity results, if any
            where = self.embed(similarity, 0)

        # Save where
        query["where"] = where

        # Run query
        return self.query(query, limit, parameters, indexids)

    def parse(self, query):
        """
        Parses a query into query components.

        Args:
            query: input query

        Returns:
            dict of parsed query components
        """

        return self.sql(query)

    def resolve(self, name, alias=None):
        """
        Resolves a query column name with the database column name. This method also builds alias expressions
        if alias is set.

        Args:
            name: query column name
            alias: alias name, defaults to None

        Returns:
            database column name
        """

        raise NotImplementedError

    def embed(self, similarity, batch):
        """
        Embeds similarity query results into a database query.

        Args:
            similarity: similarity results as [(indexid, score)]
            batch: batch id
        """

        raise NotImplementedError

    def query(self, query, limit, parameters, indexids):
        """
        Executes query against database.

        Args:
            query: input query
            limit: maximum number of results to return
            parameters: dict of named parameters to bind to placeholders
            indexids: results are returned as [(indexid, score)] regardless of select clause parameters if True

        Returns:
            query results
        """

        raise NotImplementedError

    def configure(self, config):
        """
        Initialize configuration.

        Args:
            config: configuration
        """

        # Database configuration
        self.config = config

        # SQL parser
        self.sql = SQL(self)

        # Load objects encoder
        encoder = self.config.get("objects")
        self.encoder = EncoderFactory.create(encoder) if encoder else None

        # Transform columns
        columns = config.get("columns", {})
        self.text = columns.get("text", "text")
        self.object = columns.get("object", "object")

        # Custom functions and expressions
        self.functions, self.expressions = None, None

        # Load custom functions
        self.registerfunctions(self.config)

        # Load custom expressions
        self.registerexpressions(self.config)

    def registerfunctions(self, config):
        """
        Register custom functions. This method stores the function details for underlying
        database implementations to handle.

        Args:
            config: database configuration
        """

        inputs = config.get("functions") if config else None
        if inputs:
            functions = []
            for fn in inputs:
                name, argcount = None, -1

                # Optional function configuration
                if isinstance(fn, dict):
                    name, argcount, fn = fn.get("name"), fn.get("argcount", -1), fn["function"]

                # Determine if this is a callable object or a function
                if not isinstance(fn, types.FunctionType) and hasattr(fn, "__call__"):
                    name = name if name else fn.__class__.__name__.lower()
                    fn = fn.__call__
                else:
                    name = name if name else fn.__name__.lower()

                # Store function details
                functions.append((name, argcount, fn))

            # pylint: disable=W0201
            self.functions = functions

    def registerexpressions(self, config):
        """
        Register custom expressions. This method parses and resolves expressions for later use in SQL queries.

        Args:
            config: database configuration
        """

        inputs = config.get("expressions") if config else None
        if inputs:
            expressions = {}
            for entry in inputs:
                name = entry.get("name")
                expression = entry.get("expression")
                if name and expression:
                    expressions[name] = self.sql.snippet(expression)

            # pylint: disable=W0201
            self.expressions = expressions

    def execute(self, function, *args):
        """
        Executes a user query. This method has common error handling logic.

        Args:
            function: database execute function
            args: function arguments

        Returns:
            result of function(args)
        """

        try:
            # Debug log SQL
            logger.debug(" ".join(["%s"] * len(args)), *args)

            return function(*args)
        except Exception as e:
            raise SQLError(e) from None

    def setting(self, name, default=None):
        """
        Looks up database specific setting.

        Args:
            name: setting name
            default: default value when setting not found

        Returns:
            setting value
        """

        # Get the database-specific config object
        database = self.config.get(self.config["content"])

        # Get setting value, set default value if not found
        setting = database.get(name) if database else None
        return setting if setting else default



"""
SerializeEncoder module
"""

from ...serialize import SerializeFactory

from .base import Encoder


class SerializeEncoder(Encoder):
    """
    Encodes and decodes objects using the internal serialize package.
    """

    def __init__(self, method):
        # Parent constructor
        super().__init__()

        # Pickle serialization
        self.serializer = SerializeFactory.create(method)

    def encode(self, obj):
        # Pickle object
        return self.serializer.savebytes(obj)

    def decode(self, data):
        # Unpickle to object
        return self.serializer.loadbytes(data)



"""
Factory module
"""

import json
import os

from huggingface_hub.errors import HFValidationError
from transformers.utils import cached_file

from .base import Pooling
from .cls import ClsPooling
from .late import LatePooling
from .mean import MeanPooling


class PoolingFactory:
    """
    Method to create pooling models.
    """

    @staticmethod
    def create(config):
        """
        Create a Pooling model.

        Args:
            config: pooling configuration

        Returns:
            Pooling
        """

        # Unpack parameters
        method, path, device, tokenizer, maxlength, modelargs = [
            config.get(x) for x in ["method", "path", "device", "tokenizer", "maxlength", "modelargs"]
        ]

        # Derive maxlength, if applicable
        maxlength = PoolingFactory.maxlength(path) if isinstance(maxlength, bool) and maxlength else maxlength

        # Default pooling returns hidden state
        if isinstance(path, bytes) or (isinstance(path, str) and os.path.isfile(path)) or method == "pooling":
            return Pooling(path, device, tokenizer, maxlength, modelargs)

        # Derive pooling method if it's not specified and path is a string
        if (not method or method not in ("clspooling", "meanpooling", "latepooling")) and isinstance(path, str):
            method = PoolingFactory.method(path)

        # Check for cls pooling
        if method == "clspooling":
            return ClsPooling(path, device, tokenizer, maxlength, modelargs)

        # Check for late pooling
        if method == "latepooling":
            return LatePooling(path, device, tokenizer, maxlength, modelargs)

        # Default to mean pooling
        return MeanPooling(path, device, tokenizer, maxlength, modelargs)

    @staticmethod
    def method(path):
        """
        Determines the pooling method using the sentence transformers pooling config.

        Args:
            path: model path

        Returns:
            pooling method
        """

        # Default method
        method = "meanpooling"

        # Load 1_Pooling/config.json file
        config = PoolingFactory.load(path, "1_Pooling/config.json")

        # Set to CLS pooling if it's enabled and mean pooling is disabled
        if config and config["pooling_mode_cls_token"] and not config["pooling_mode_mean_tokens"]:
            method = "clspooling"

        # Check for late interaction pooling
        if not config:
            # Load 1_Dense/config.json
            config = PoolingFactory.load(path, "1_Dense/config.json")
            if config:
                method = "latepooling"

            # Load config.json and check architecture
            else:
                config = PoolingFactory.load(path, "config.json")
                if config and "HF_ColBERT" in config.get("architectures", []):
                    method = "latepooling"

        return method

    @staticmethod
    def maxlength(path):
        """
        Reads the max_seq_length parameter from sentence transformers config.

        Args:
            path: model path

        Returns:
            max sequence length
        """

        # Default length is unset
        maxlength = None

        # Read max_seq_length from sentence_bert_config.json
        config = PoolingFactory.load(path, "sentence_bert_config.json")
        maxlength = config.get("max_seq_length") if config else maxlength

        return maxlength

    @staticmethod
    def load(path, name):
        """
        Loads a JSON config file from the Hugging Face Hub.

        Args:
            path: model path
            name: file to load

        Returns:
            config
        """

        # Download file and parse JSON
        config = None
        try:
            path = cached_file(path_or_repo_id=path, filename=name)
            if path:
                with open(path, encoding="utf-8") as f:
                    config = json.load(f)

        # Ignore this error - invalid repo or directory
        except (HFValidationError, OSError):
            pass

        return config



"""
FastAPI application module
"""

import inspect
import os
import sys

from fastapi import APIRouter, Depends, FastAPI
from fastapi_mcp import FastApiMCP
from httpx import AsyncClient

from .authorization import Authorization
from .base import API
from .factory import APIFactory

from ..app import Application


def get():
    """
    Returns global API instance.

    Returns:
        API instance
    """

    return INSTANCE


def create():
    """
    Creates a FastAPI instance.
    """

    # Application dependencies
    dependencies = []

    # Default implementation of token authorization
    token = os.environ.get("TOKEN")
    if token:
        dependencies.append(Depends(Authorization(token)))

    # Add custom dependencies
    deps = os.environ.get("DEPENDENCIES")
    if deps:
        for dep in deps.split(","):
            # Create and add dependency
            dep = APIFactory.get(dep.strip())()
            dependencies.append(Depends(dep))

    # Create FastAPI application
    return FastAPI(lifespan=lifespan, dependencies=dependencies if dependencies else None)


def apirouters():
    """
    Lists available APIRouters.

    Returns:
        {router name: router}
    """

    # Get handle to api module
    api = sys.modules[".".join(__name__.split(".")[:-1])]

    available = {}
    for name, rclass in inspect.getmembers(api, inspect.ismodule):
        if hasattr(rclass, "router") and isinstance(rclass.router, APIRouter):
            available[name.lower()] = rclass.router

    return available


def lifespan(application):
    """
    FastAPI lifespan event handler.

    Args:
        application: FastAPI application to initialize
    """

    # pylint: disable=W0603
    global INSTANCE

    # Load YAML settings
    config = Application.read(os.environ.get("CONFIG"))

    # Instantiate API instance
    api = os.environ.get("API_CLASS")
    INSTANCE = APIFactory.create(config, api) if api else API(config)

    # Get all known routers
    routers = apirouters()

    # Conditionally add routes based on configuration
    for name, router in routers.items():
        if name in config:
            application.include_router(router)

    # Special case for embeddings clusters
    if "cluster" in config and "embeddings" not in config:
        application.include_router(routers["embeddings"])

    # Special case to add similarity instance for embeddings
    if "embeddings" in config and "similarity" not in config:
        application.include_router(routers["similarity"])

    # Execute extensions if present
    extensions = os.environ.get("EXTENSIONS")
    if extensions:
        for extension in extensions.split(","):
            # Create instance and execute extension
            extension = APIFactory.get(extension.strip())()
            extension(application)

    # Add Model Context Protocol (MCP) service, if applicable
    if config.get("mcp"):
        mcp = FastApiMCP(application, http_client=AsyncClient(timeout=100))
        mcp.mount()

    yield


def start():
    """
    Runs application lifespan handler.
    """

    list(lifespan(app))


# FastAPI instance txtai API instances
app, INSTANCE = create(), None



"""
Textractor module
"""

import contextlib
import os
import tempfile

from urllib.parse import urlparse
from urllib.request import urlopen, Request

from .filetohtml import FileToHTML
from .htmltomd import HTMLToMarkdown
from .segmentation import Segmentation


class Textractor(Segmentation):
    """
    Extracts text from files.
    """

    # pylint: disable=R0913
    def __init__(
        self,
        sentences=False,
        lines=False,
        paragraphs=False,
        minlength=None,
        join=False,
        sections=False,
        cleantext=True,
        chunker=None,
        headers=None,
        backend="available",
        **kwargs
    ):
        super().__init__(sentences, lines, paragraphs, minlength, join, sections, cleantext, chunker, **kwargs)

        # Get backend parameter - handle legacy tika flag
        backend = "tika" if "tika" in kwargs and kwargs["tika"] else None if "tika" in kwargs else backend

        # File to HTML pipeline
        self.html = FileToHTML(backend) if backend else None

        # HTML to Markdown pipeline
        self.markdown = HTMLToMarkdown(self.paragraphs, self.sections)

        # HTTP headers
        self.headers = headers if headers else {}

    def text(self, text):
        # Check if text is a valid file path or url
        path, exists = self.valid(text)

        if not path:
            # Not a valid file path, treat input as data
            html = text

        elif self.html:
            # Use FileToHTML pipeline, if available
            # Retrieve remote file, if necessary
            path = path if exists else self.download(path)

            # Parse content to HTML
            html = self.html(path)

            # FiletoHTML pipeline returns None when input is already HTML
            html = html if html else self.retrieve(path)

            # Delete temporary file
            if not exists:
                os.remove(path)

        else:
            # Read data from url/path
            html = self.retrieve(path)

        # HTML to Markdown
        return self.markdown(html)

    def valid(self, path):
        """
        Checks if path is a valid local file or web url. Returns path if valid along with a flag
        denoting if the path exists locally.

        Args:
            path: path to check

        Returns:
            (path, exists)
        """

        # Convert file urls to local paths
        path = path.replace("file://", "")

        # Check if this is a local file path or local file url
        exists = os.path.exists(path)

        # Consider local files and HTTP urls valid
        return (path if exists or urlparse(path).scheme in ("http", "https") else None, exists)

    def download(self, url):
        """
        Downloads content of url to a temporary file.

        Args:
            url: input url

        Returns:
            temporary file path
        """

        with tempfile.NamedTemporaryFile(mode="wb", delete=False) as output:
            path = output.name

            # Retrieve and write data to temporary file
            output.write(self.retrieve(url))

        return path

    def retrieve(self, url):
        """
        Retrieves content from url.

        Args:
            url: input url

        Returns:
            data
        """

        # Local file
        if os.path.exists(url):
            with open(url, "rb") as f:
                return f.read()

        # Remote file
        with contextlib.closing(urlopen(Request(url, headers=self.headers))) as connection:
            return connection.read()



"""
Zip module
"""

import os

from zipfile import ZipFile, ZIP_DEFLATED

from .compress import Compress


class Zip(Compress):
    """
    Zip compression
    """

    def pack(self, path, output):
        with ZipFile(output, "w", ZIP_DEFLATED) as zfile:
            for root, _, files in sorted(os.walk(path)):
                for f in files:
                    # Generate archive name with relative path, if necessary
                    name = os.path.join(os.path.relpath(root, path), f)

                    # Write file to zip
                    zfile.write(os.path.join(root, f), arcname=name)

    def unpack(self, path, output):
        with ZipFile(path, "r") as zfile:
            # Validate path if directory specified
            for fullpath in zfile.namelist():
                fullpath = os.path.join(path, fullpath)
                if os.path.dirname(fullpath) and not self.validate(path, fullpath):
                    raise IOError(f"Invalid zip entry: {fullpath}")

            zfile.extractall(output)



"""
RDBMS module
"""

import os

# Conditional import
try:
    from grand import Graph
    from grand.backends import SQLBackend, InMemoryCachedBackend

    from sqlalchemy import create_engine, text, StaticPool
    from sqlalchemy.schema import CreateSchema

    ORM = True
except ImportError:
    ORM = False

from .networkx import NetworkX


class RDBMS(NetworkX):
    """
    Graph instance backed by a relational database.
    """

    def __init__(self, config):
        # Check before super() in case those required libraries are also not available
        if not ORM:
            raise ImportError('RDBMS is not available - install "graph" extra to enable')

        super().__init__(config)

        # Graph and database instances
        self.graph = None
        self.database = None

    def __del__(self):
        if hasattr(self, "database") and self.database:
            self.database.close()

    def create(self):
        # Create graph instance
        self.graph, self.database = self.connect()

        # Clear previous graph, if available
        for table in [self.config.get("nodes", "nodes"), self.config.get("edges", "edges")]:
            self.database.execute(text(f"DELETE FROM {table}"))

        # Return NetworkX compatible backend
        return self.graph.nx

    def scan(self, attribute=None, data=False):
        if attribute:
            for node in self.backend:
                attributes = self.node(node)
                if attribute in attributes:
                    yield (node, attributes) if data else node
        else:
            yield from super().scan(attribute, data)

    def load(self, path):
        # Create graph instance
        self.graph, self.database = self.connect()

        # Store NetworkX compatible backend
        self.backend = self.graph.nx

    def save(self, path):
        self.database.commit()

    def close(self):
        # Parent logic
        super().close()

        # Close database connection
        self.database.close()

    def filter(self, nodes, graph=None):
        return super().filter(nodes, graph if graph else NetworkX(self.config))

    def connect(self):
        """
        Connects to a graph backed by a relational database.

        Args:
            Graph database instance
        """

        # Keyword arguments for SQLAlchemy
        kwargs = {"poolclass": StaticPool, "echo": False}
        url = self.config.get("url", os.environ.get("GRAPH_URL"))

        # Set default schema, if necessary
        schema = self.config.get("schema")
        if schema:
            # Check that schema exists
            engine = create_engine(url)
            with engine.begin() as connection:
                connection.execute(CreateSchema(schema, if_not_exists=True) if "postgresql" in url else text("SELECT 1"))

            # Set default schema
            kwargs["connect_args"] = {"options": f'-c search_path="{schema}"'} if "postgresql" in url else {}

        backend = SQLBackend(
            db_url=url,
            node_table_name=self.config.get("nodes", "nodes"),
            edge_table_name=self.config.get("edges", "edges"),
            sqlalchemy_kwargs=kwargs,
        )

        # pylint: disable=W0212
        return Graph(backend=InMemoryCachedBackend(backend, maxsize=None)), backend._connection



"""
Encoder module
"""

from io import BytesIO


class Encoder:
    """
    Encodes and decodes object content. The base encoder works only with byte arrays. It can be extended to encode different datatypes.
    """

    def encode(self, obj):
        """
        Encodes an object to a byte array using the encoder.

        Args:
            obj: object to encode

        Returns:
            encoded object as a byte array
        """

        return obj

    def decode(self, data):
        """
        Decodes input byte array into an object using this encoder.

        Args:
            data: encoded data

        Returns:
            decoded object
        """

        return BytesIO(data) if data else None



"""
StreamTask module
"""

from .base import Task


class StreamTask(Task):
    """
    Task that calls a task action and yields results.
    """

    def register(self, batch=False):
        """
        Adds stream parameters to task.

        Args:
            batch: all elements are passed to a single action call if True, otherwise an action call is executed per element, defaults to False
        """

        # pylint: disable=W0201
        # All elements are passed to a single action call if True, otherwise an action call is executed per element, defaults to False
        self.batch = batch

    def __call__(self, elements, executor=None):
        for action in self.action:
            if self.batch:
                # Single batch call
                yield from action(elements)
            else:
                # Call action for each element
                for x in elements:
                    yield from action(x)



"""
Tabular module
"""

import os

# Conditional import
try:
    import pandas as pd

    PANDAS = True
except ImportError:
    PANDAS = False

from ..base import Pipeline


class Tabular(Pipeline):
    """
    Splits tabular data into rows and columns.
    """

    def __init__(self, idcolumn=None, textcolumns=None, content=False):
        """
        Creates a new Tabular pipeline.

        Args:
            idcolumn: column name to use for row id
            textcolumns: list of columns to combine as a text field
            content: if True, a dict per row is generated with all fields. If content is a list, a subset of fields
                     is included in the generated rows.
        """

        if not PANDAS:
            raise ImportError('Tabular pipeline is not available - install "pipeline" extra to enable')

        self.idcolumn = idcolumn
        self.textcolumns = textcolumns
        self.content = content

    def __call__(self, data):
        """
        Splits data into rows and columns.

        Args:
            data: input data

        Returns:
            list of (id, text, tag)
        """

        items = [data] if not isinstance(data, list) else data

        # Combine all rows into single return element
        results = []
        dicts = []

        for item in items:
            # File path
            if isinstance(item, str):
                _, extension = os.path.splitext(item)
                extension = extension.replace(".", "").lower()

                if extension == "csv":
                    df = pd.read_csv(item)

                results.append(self.process(df))

            # Dict
            if isinstance(item, dict):
                dicts.append(item)

            # List of dicts
            elif isinstance(item, list):
                df = pd.DataFrame(item)
                results.append(self.process(df))

        if dicts:
            df = pd.DataFrame(dicts)
            results.extend(self.process(df))

        return results[0] if not isinstance(data, list) else results

    def process(self, df):
        """
        Extracts a list of (id, text, tag) tuples from a dataframe.

        Args:
            df: DataFrame to extract content from

        Returns:
            list of (id, text, tag)
        """

        rows = []

        # Columns to use for text
        columns = self.textcolumns
        if not columns:
            columns = list(df.columns)
            if self.idcolumn:
                columns.remove(self.idcolumn)

        # Transform into (id, text, tag) tuples
        for index, row in df.iterrows():
            uid = row[self.idcolumn] if self.idcolumn else index
            uid = uid if uid is not None else index
            text = self.concat(row, columns)

            rows.append((uid, text, None))

            # Also add row for content
            if isinstance(self.content, list):
                row = {column: self.column(value) for column, value in row.to_dict().items() if column in self.content}
                rows.append((uid, row, None))
            elif self.content:
                row = {column: self.column(value) for column, value in row.to_dict().items()}
                rows.append((uid, row, None))

        return rows

    def concat(self, row, columns):
        """
        Builds a text field from row using columns.

        Args:
            row: input row
            columns: list of columns to join together

        Returns:
            text
        """

        parts = []
        for column in columns:
            column = self.column(row[column])
            if column:
                parts.append(str(column))

        return ". ".join(parts) if parts else None

    def column(self, value):
        """
        Applies column standardization logic:
            - Replace NaN values with None

        Args:
            value: input value

        Returns:
            formatted value
        """

        # Check for null - treat lists as not null
        return None if not isinstance(value, list) and pd.isnull(value) else value



"""
AudioMixer module
"""

from ..base import Pipeline
from .signal import Signal, SCIPY


class AudioMixer(Pipeline):
    """
    Mixes multiple audio streams into a single stream.
    """

    def __init__(self, rate=None):
        """
        Creates an AudioMixer pipeline.

        Args:
            rate: optional target sample rate, otherwise uses input target rate with each audio segment
        """

        if not SCIPY:
            raise ImportError('AudioMixer pipeline is not available - install "pipeline" extra to enable.')

        # Target sample rate
        self.rate = rate

    def __call__(self, segment, scale1=1, scale2=1):
        """
        Mixes multiple audio streams into a single stream.

        Args:
            segment: ((audio1, sample rate), (audio2, sample rate))|list
            scale1: optional scaling factor for segment1
            scale2: optional scaling factor for segment2

        Returns:
            list of (audio, sample rate)
        """

        # Convert single element to list
        segments = [segment] if isinstance(segment, tuple) else segment

        results = []
        for segment1, segment2 in segments:
            audio1, rate1 = segment1
            audio2, rate2 = segment2

            # Resample audio, as necessary
            target = self.rate if self.rate else rate1
            audio1 = Signal.resample(audio1, rate1, target)
            audio2 = Signal.resample(audio2, rate2, target)

            # Mix audio into single segment
            results.append((Signal.mix(audio1, audio2, scale1, scale2), target))

        # Return single element if single element passed in
        return results[0] if isinstance(segment, tuple) else results



"""
SparseArray module
"""

import numpy as np

# Conditional import
try:
    from scipy.sparse import csr_matrix

    SCIPY = True
except ImportError:
    SCIPY = False


class SparseArray:
    """
    Methods to load and save sparse arrays to file.
    """

    def __init__(self):
        """
        Creates a SparseArray instance.
        """

        if not SCIPY:
            raise ImportError("SciPy is not available - install scipy to enable")

    def load(self, f):
        """
        Loads a sparse array from file.

        Args:
            f: input file handle

        Returns:
            sparse array
        """

        # Load raw data
        data, indices, indptr, shape = (
            np.load(f, allow_pickle=False),
            np.load(f, allow_pickle=False),
            np.load(f, allow_pickle=False),
            np.load(f, allow_pickle=False),
        )

        # Load data into sparse array
        return csr_matrix((data, indices, indptr), shape=shape)

    def save(self, f, array):
        """
        Saves a sparse array to file.

        Args:
            f: output file handle
            array: sparse array
        """

        # Save sparse array to file
        for x in [array.data, array.indices, array.indptr, array.shape]:
            np.save(f, x, allow_pickle=False)



"""
Audio imports
"""

from .audiomixer import AudioMixer
from .audiostream import AudioStream
from .microphone import Microphone
from .signal import Signal
from .texttoaudio import TextToAudio
from .texttospeech import TextToSpeech
from .transcription import Transcription



"""
FileToHTML module
"""

import os
import re

from subprocess import Popen

# Conditional import
try:
    from tika import detector, parser

    TIKA = True
except ImportError:
    TIKA = False

# Conditional import
try:
    from docling.document_converter import DocumentConverter

    DOCLING = True
except ImportError:
    DOCLING = False

from ..base import Pipeline


class FileToHTML(Pipeline):
    """
    File to HTML pipeline.
    """

    def __init__(self, backend="available"):
        """
        Creates a new File to HTML pipeline.

        Args:
            backend: backend to use to extract content, supports "tika", "docling" or "available" (default) which finds the first available
        """

        # Lowercase backend parameter
        backend = backend.lower() if backend else None

        # Check for available backend
        if backend == "available":
            backend = "tika" if Tika.available() else "docling" if Docling.available() else None

        # Create backend instance
        self.backend = Tika() if backend == "tika" else Docling() if backend == "docling" else None

    def __call__(self, path):
        """
        Converts file at path to HTML. Returns None if no backend is available.

        Args:
            path: input file path

        Returns:
            html if a backend is available, otherwise returns None
        """

        return self.backend(path) if self.backend else None


class Tika:
    """
    File to HTML conversion via Apache Tika.
    """

    @staticmethod
    def available():
        """
        Checks if a Java executable is available and Tika is installed.

        Returns:
            True if Java is available and Tika is installed, False otherwise
        """

        # Get path to Java executable
        path = os.environ.get("TIKA_JAVA", "java")

        # pylint: disable=R1732,W0702,W1514
        # Check if Java binary is available on path
        try:
            _ = Popen(path, stdout=open(os.devnull, "w"), stderr=open(os.devnull, "w"))
        except:
            return False

        # Return True if Java is available AND Tika is installed
        return TIKA

    def __init__(self):
        """
        Creates a new Tika instance.
        """

        if not Tika.available():
            raise ImportError('Tika engine is not available - install "pipeline" extra to enable. Also check that Java is available.')

    def __call__(self, path):
        """
        Parses content to HTML.

        Args:
            path: file path

        Returns:
            html
        """

        # Skip parsing if input is plain text or HTML
        mimetype = detector.from_file(path)
        if mimetype in ("text/plain", "text/html", "text/xhtml"):
            return None

        # Parse content to HTML
        parsed = parser.from_file(path, xmlContent=True)
        return parsed["content"]


class Docling:
    """
    File to HTML conversion via Docling.
    """

    @staticmethod
    def available():
        """
        Checks if Docling is available.

        Returns:
            True if Docling is available, False otherwise
        """

        return DOCLING

    def __init__(self):
        """
        Creates a new Docling instance.
        """

        if not Docling.available():
            raise ImportError('Docling engine is not available - install "pipeline" extra to enable')

        self.converter = DocumentConverter()

    def __call__(self, path):
        """
        Parses content to HTML.

        Args:
            path: file path

        Returns:
            html
        """

        # Skip parsing if input is HTML
        if self.ishtml(path):
            return None

        # Parse content to HTML
        html = self.converter.convert(path).document.export_to_html(html_head="<head/>")

        # Normalize HTML and return
        return self.normalize(html)

    def ishtml(self, path):
        """
        Detects if this file looks like HTML.

        Args:
            path: file path

        Returns:
            True if this is HTML
        """

        with open(path, "rb") as f:
            # Read first 1024 bytes, ignore encoding errors and strip leading/trailing whitespace
            content = f.read(1024)
            content = content.decode("ascii", errors="ignore").lower().strip()

            # Check for HTML
            return re.search(r"<!doctype\s+html|<html|<head|<body", content)

    def normalize(self, html):
        """
        Applies normalization rules to make HTML consistent with other text extraction backends.

        Args:
            html: input html

        Returns:
            normalized html
        """

        # Wrap content with a body tag, if necessary
        html = html.replace("<head/>", "<head/><body>").replace("</html>", "</body></html>") if "<body>" not in html else html

        # Remove bullets from list items
        html = re.sub(r"<li>\xb7 ", r"<li>", html)

        # Add spacing between paragraphs
        return html.replace("</p>", "</p><p/>")



"""
ExportTask module
"""

import datetime
import os

# Conditional import
try:
    import pandas as pd

    PANDAS = True
except ImportError:
    PANDAS = False

from .base import Task


class ExportTask(Task):
    """
    Task that exports task elements using Pandas.
    """

    def register(self, output=None, timestamp=None):
        """
        Add export parameters to task. Checks if required dependencies are installed.

        Args:
            output: output file path
            timestamp: true if output file should be timestamped
        """

        if not PANDAS:
            raise ImportError('ExportTask is not available - install "workflow" extra to enable')

        # pylint: disable=W0201
        self.output = output
        self.timestamp = timestamp

    def __call__(self, elements, executor=None):
        # Run task
        outputs = super().__call__(elements, executor)

        # Get output file extension
        output = self.output
        parts = list(os.path.splitext(output))
        extension = parts[-1].lower()

        # Add timestamp to filename
        if self.timestamp:
            timestamp = datetime.datetime.now(datetime.timezone.utc).strftime("%Y%m%dT%H%M%SZ")
            parts[-1] = timestamp + parts[-1]

            # Create full path to output file
            output = ".".join(parts)

        # Write output
        if extension == ".xlsx":
            pd.DataFrame(outputs).to_excel(output, index=False)
        else:
            pd.DataFrame(outputs).to_csv(output, index=False)

        # Return results
        return outputs



"""
Reranker module
"""

from ..base import Pipeline


class Reranker(Pipeline):
    """
    Runs embeddings queries and re-ranks them using a similarity pipeline. Note that content must be enabled with the
    embeddings instance for this to work properly.
    """

    def __init__(self, embeddings, similarity):
        """
        Creates a Reranker pipeline.

        Args:
            embeddings: embeddings instance (content must be enabled)
            similarity: similarity instance
        """

        self.embeddings, self.similarity = embeddings, similarity

    # pylint: disable=W0222
    def __call__(self, query, limit=3, factor=10, **kwargs):
        """
        Runs an embeddings search and re-ranks the results using a Similarity pipeline.

        Args:
            query: query text|list
            limit: maximum results
            factor: factor to multiply limit by for the initial embeddings search
            kwargs: additional arguments to pass to embeddings search

        Returns:
            list of query results rescored using a Similarity pipeline
        """

        queries = [query] if not isinstance(query, list) else query

        # Run searches
        results = self.embeddings.batchsearch(queries, limit * factor, **kwargs)

        # Re-rank using similarity pipeline
        ranked = []
        for x, result in enumerate(results):
            texts = [row["text"] for row in result]

            # Score results and merge
            for uid, score in self.similarity(queries[x], texts):
                result[uid]["score"] = score

            # Sort and take top n sorted results
            ranked.append(sorted(result, key=lambda row: row["score"], reverse=True)[:limit])

        return ranked[0] if isinstance(query, str) else ranked



"""
Defines an OpenAI-compatible API endpoint for txtai.

See the following specification for more information:
https://github.com/openai/openai-openapi
"""

import uuid
import json
import time

from typing import List, Optional, Union

from fastapi import APIRouter, Body, Form, UploadFile
from fastapi.responses import Response, StreamingResponse

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


# pylint: disable=W0622
@router.post("/v1/chat/completions")
def chat(
    messages: List[dict] = Body(...),
    model: str = Body(...),
    max_completion_tokens: Optional[int] = Body(default=None),
    stream: Optional[bool] = Body(default=False),
):
    """
    Runs a chat completion request.

    Args:
        messages: list of messages [{"role": role, "content": content}]
        model: agent name, workflow name, pipeline name or embeddings
        max_completion_tokens: sets the max length to generate
        stream: streams response if True

    Returns:
        chat completion
    """

    # Build keyword arguments
    kwargs = {key: value for key, value in [("stream", stream), ("maxlength", max_completion_tokens)] if value}

    # Get first message
    message = messages[0]["content"]

    # Agent
    if model in application.get().agents:
        result = application.get().agent(model, message, **kwargs)

    # Embeddings search
    elif model == "embeddings":
        result = application.get().search(message, 1, **kwargs)[0]["text"]

    # Pipeline
    elif model in application.get().pipelines and model != "llm":
        result = application.get().pipeline(model, message, **kwargs)

    # Workflow
    elif model in application.get().workflows:
        result = list(application.get().workflow(model, [message], **kwargs))[0]

    # Default to running all messages through default LLM
    else:
        result = application.get().pipeline("llm", messages, **kwargs)

    # Write response
    return StreamingResponse(StreamingChatResponse()(model, result)) if stream else ChatResponse()(model, result)


@router.post("/v1/embeddings")
def embeddings(input: Union[str, List[str]] = Body(...), model: str = Body(...)):
    """
    Creates an embeddings vector for the input text.

    Args:
        input: text|list
        model: model name

    Returns:
        list of embeddings vectors
    """

    # Convert to embeddings
    result = application.get().batchtransform([input] if isinstance(input, str) else input)

    # Build and return response
    data = []
    for index, embedding in enumerate(result):
        data.append({"object": "embedding", "embedding": embedding, "index": index})

    return {"object": "list", "data": data, "model": model}


@router.post("/v1/audio/speech")
def speech(input: str = Body(...), voice: str = Body(...), response_format: Optional[str] = Body(default="mp3")):
    """
    Generates speech for the input text.

    Args:
        input: input text
        voice: speaker name
        response_format: audio encoding format, defaults to mp3

    Returns:
        audio data
    """

    # Convert to audio
    audio = application.get().pipeline("texttospeech", input, speaker=voice, encoding=response_format)

    # Write audio
    return Response(audio)


@router.post("/v1/audio/transcriptions")
def transcribe(file: UploadFile, language: Optional[str] = Form(default=None), response_format: Optional[str] = Form(default="json")):
    """
    Transcribes audio to text.

    Args:
        file: audio input file
        language: language of input audio
        response_format: output format (json or text)

    Returns:
        transcribed text
    """

    # Transcribe
    text = application.get().pipeline("transcription", file.file, language=language, task="transcribe")
    return text if response_format == "text" else {"text": text}


@router.post("/v1/audio/translations")
def translate(
    file: UploadFile,
    response_format: Optional[str] = Form(default="json"),
):
    """
    Translates audio to English.

    Args:
        file: audio input file
        response_format: output format (json or text)

    Returns:
        translated text
    """

    # Transcribe and translate to English
    text = application.get().pipeline("transcription", file.file, language="English", task="translate")
    return text if response_format == "text" else {"text": text}


class ChatResponse:
    """
    Returns a chat response object.
    """

    def __call__(self, model, result):
        return {
            "id": str(uuid.uuid4()),
            "object": "chat.completion",
            "created": int(time.time() * 1000),
            "model": model,
            "choices": [{"id": 0, "message": {"role": "assistant", "content": result}, "finish_reason": "stop"}],
        }


class StreamingChatResponse:
    """
    Returns a streaming chat response object.
    """

    def __call__(self, model, result):
        for chunk in result:
            yield "data: " + json.dumps(
                {
                    "id": str(uuid.uuid4()),
                    "object": "chat.completion.chunk",
                    "created": int(time.time() * 1000),
                    "model": model,
                    "choices": [{"id": 0, "delta": {"content": chunk}}],
                }
            ) + "\n\n"

        yield "data: [DONE]\n\n"



"""
Segment imports
"""

from .filetohtml import FileToHTML
from .htmltomd import HTMLToMarkdown
from .segmentation import Segmentation
from .tabular import Tabular
from .textractor import Textractor
from .tokenizer import Tokenizer



# pylint: disable = C0111
from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    # Remove GitHub dark mode images
    DESCRIPTION = "".join([line for line in f if "gh-dark-mode-only" not in line])

# Required dependencies
install = ["faiss-cpu>=1.7.1.post2", "msgpack>=1.0.7", "torch>=2.1", "transformers>=4.45.0"]

# Required dependencies that are also base transformers dependencies
install += ["huggingface-hub>=0.34.0", "numpy>=1.18.4", "pyyaml>=5.3", "regex>=2022.8.17", "safetensors>=0.4.5"]

# Optional dependencies
extras = {}

# Development dependencies - not included in "all" install
extras["dev"] = [
    "black",
    "coverage",
    "coveralls",
    "httpx",
    "mkdocs-material",
    "mkdocs-redirects",
    "mkdocstrings[python]",
    "pre-commit",
    "pylint",
]

extras["agent"] = ["mcpadapt>=0.1.0", "smolagents>=1.17"]

extras["ann"] = [
    "annoy>=1.16.3",
    "scikit-learn>=0.23.1",
    "hnswlib>=0.5.0",
    "pgvector>=0.4.1",
    "scipy>=1.4.1",
    "sqlalchemy>=2.0.20",
    "sqlite-vec>=0.1.1",
]

extras["api"] = [
    "aiohttp>=3.8.1",
    "fastapi>=0.94.0",
    "fastapi-mcp>=0.2.0",
    "httpx>=0.28.1",
    "pillow>=7.1.2",
    "python-multipart>=0.0.7",
    "uvicorn>=0.12.1",
]

extras["cloud"] = ["apache-libcloud>=3.3.1", "fasteners>=0.14.1"]

extras["console"] = ["rich>=12.0.1"]

extras["database"] = ["duckdb>=0.7.1", "pillow>=7.1.2", "sqlalchemy>=2.0.20"]

extras["graph"] = ["grand-cypher>=0.6.0", "grand-graph>=0.6.0", "networkx>=2.7.1", "sqlalchemy>=2.0.20"]

extras["model"] = ["onnx>=1.11.0", "onnxruntime>=1.11.0"]

extras["pipeline-audio"] = [
    "onnx>=1.11.0",
    "onnxruntime>=1.11.0",
    "scipy>=1.4.1",
    "sounddevice>=0.5.0",
    "soundfile>=0.10.3.post1",
    "ttstokenizer>=1.1.0",
    "webrtcvad-wheels>=2.0.14",
]

extras["pipeline-data"] = ["beautifulsoup4>=4.9.3", "chonkie>=1.0.2", "docling>=2.8.2", "nltk>=3.5", "pandas>=1.1.0", "tika>=1.24"]

extras["pipeline-image"] = ["imagehash>=4.2.1", "pillow>=7.1.2", "timm>=0.4.12"]

extras["pipeline-llm"] = ["litellm>=1.37.16", "llama-cpp-python>=0.2.75"]

extras["pipeline-text"] = ["gliner>=0.2.16", "sentencepiece>=0.1.91", "staticvectors>=0.2.0"]

extras["pipeline-train"] = [
    "accelerate>=0.26.0",
    "bitsandbytes>=0.42.0",
    "onnx>=1.11.0",
    "onnxmltools>=1.9.1",
    "onnxruntime>=1.11.0",
    "peft>=0.8.1",
    "skl2onnx>=1.9.1",
]

extras["pipeline"] = (
    extras["pipeline-audio"]
    + extras["pipeline-data"]
    + extras["pipeline-image"]
    + extras["pipeline-llm"]
    + extras["pipeline-text"]
    + extras["pipeline-train"]
)

extras["scoring"] = ["sqlalchemy>=2.0.20"]

extras["vectors"] = [
    "litellm>=1.37.16",
    "llama-cpp-python>=0.2.75",
    "model2vec>=0.3.0",
    "scikit-learn>=0.23.1",
    "scipy>=1.4.1",
    "sentence-transformers>=5.0.0",
    "skops>=0.9.0",
    "staticvectors>=0.2.0",
]

extras["workflow"] = [
    "apache-libcloud>=3.3.1",
    "croniter>=1.2.0",
    "openpyxl>=3.0.9",
    "pandas>=1.1.0",
    "pillow>=7.1.2",
    "requests>=2.26.0",
    "xmltodict>=0.12.0",
]

# Backwards-compatible combination of ann and vectors extra
extras["similarity"] = extras["ann"] + extras["vectors"]

extras["all"] = (
    extras["agent"]
    + extras["api"]
    + extras["cloud"]
    + extras["console"]
    + extras["database"]
    + extras["graph"]
    + extras["model"]
    + extras["pipeline"]
    + extras["scoring"]
    + extras["similarity"]
    + extras["workflow"]
)

setup(
    name="mseep-txtai",
    version="9.1.1",
    author="mseep",
    description="All-in-one open-source AI framework for semantic search, LLM orchestration and language model workflows",
    long_description="Package managed by MseeP.ai",
    long_description_content_type="text/plain",
    url="https://github.com/neuml/txtai",
    project_urls={
        "Documentation": "https://github.com/neuml/txtai",
        "Issue Tracker": "https://github.com/neuml/txtai/issues",
        "Source Code": "https://github.com/neuml/txtai",
    },
    license="Apache 2.0: http://www.apache.org/licenses/LICENSE-2.0",
    packages=find_packages(where="src/python"),
    package_dir={"": "src/python"},
    keywords="search embedding machine-learning nlp",
    python_requires=">=3.10",
    install_requires=install,
    extras_require=extras,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Text Processing :: Indexing",
        "Topic :: Utilities",
    ],
)



"""
Agent module
"""

from .factory import ProcessFactory


class Agent:
    """
    An agent automatically creates workflows to answer multi-faceted user requests. Agents iteratively prompt and/or interface with tools to
    step through a process and ultimately come to an answer for a request.

    Agents excel at complex tasks where multiple tools and/or methods are required. They incorporate a level of randomness similar to different
    people working on the same task. When the request is simple and/or there is a rule-based process, other methods such as RAG and Workflows
    should be explored.
    """

    def __init__(self, **kwargs):
        """
        Creates a new Agent.

        Args:
            kwargs: arguments to pass to the underlying Agent backend and LLM pipeline instance
        """

        # Ensure backwards compatibility
        if "max_iterations" in kwargs:
            kwargs["max_steps"] = kwargs.pop("max_iterations")

        # Create agent process runner
        self.process = ProcessFactory.create(kwargs)

        # Tools dictionary
        self.tools = self.process.tools

    def __call__(self, text, maxlength=8192, stream=False, **kwargs):
        """
        Runs an agent loop.

        Args:
            text: instructions to run
            maxlength: maximum sequence length
            stream: stream response if True, defaults to False
            kwargs: additional keyword arguments

        Returns:
            result
        """

        # Process parameters
        self.process.model.parameters(maxlength)

        # Run agent loop
        return self.process.run(text, stream=stream, **kwargs)



"""
Sparse Sentence Transformers module
"""

# Conditional import
try:
    from sentence_transformers import SparseEncoder

    SENTENCE_TRANSFORMERS = True
except ImportError:
    SENTENCE_TRANSFORMERS = False

from ..dense.sbert import STVectors
from .base import SparseVectors


class SparseSTVectors(SparseVectors, STVectors):
    """
    Builds sparse vectors using sentence-transformers (aka SBERT).
    """

    def __init__(self, config, scoring, models):
        # Check before parent constructor since it calls loadmodel
        if not SENTENCE_TRANSFORMERS:
            raise ImportError('sentence-transformers is not available - install "vectors" extra to enable')

        super().__init__(config, scoring, models)

    def loadencoder(self, path, device, **kwargs):
        return SparseEncoder(path, device=device, **kwargs)

    def defaultnormalize(self):
        # Enable normalization by default if similarity function is cosine
        return self.model and self.model.similarity_fn_name == "cosine"



"""
Base imports
"""

import logging

# Top-level imports
from .agent import Agent
from .app import Application
from .embeddings import Embeddings
from .pipeline import LLM, RAG
from .workflow import Workflow

# Configure logging per standard Python library recommendations
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())



"""
External module
"""

import types

import numpy as np

from ...util import Resolver

from ..base import Vectors


class External(Vectors):
    """
    Builds vectors using an external method. This can be a local function or an external API call.
    """

    def __init__(self, config, scoring, models):
        super().__init__(config, scoring, models)

        # Lookup and resolve transform function
        self.transform = self.resolve(config.get("transform"))

    def loadmodel(self, path):
        return None

    def encode(self, data, category=None):
        # Call external transform function, if available and data not already an array
        # Batching is handed by the external transform function
        if self.transform and data and not isinstance(data[0], np.ndarray):
            data = self.transform(data)

        # Cast to float32
        return data.astype(np.float32) if isinstance(data, np.ndarray) else np.array(data, dtype=np.float32)

    def resolve(self, transform):
        """
        Resolves a transform function.

        Args:
            transform: transform function

        Returns:
            resolved transform function
        """

        if transform:
            # Resolve transform instance, if necessary
            transform = Resolver()(transform) if transform and isinstance(transform, str) else transform

            # Get function or callable instance
            transform = transform if isinstance(transform, types.FunctionType) else transform()

        return transform



"""
Serializer module
"""

from .errors import SerializeError
from .factory import SerializeFactory


class Serializer:
    """
    Methods to serialize and deserialize data.
    """

    @staticmethod
    def load(path):
        """
        Loads data from path. This method first tries to load the default serialization format.
        If that fails, it will fallback to pickle format for backwards-compatability purposes.

        Note that loading pickle files requires the env variable `ALLOW_PICKLE=True`.

        Args:
            path: data to load

        Returns:
            data
        """

        try:
            return SerializeFactory.create().load(path)
        except SerializeError:
            # Backwards compatible check for pickled data
            return SerializeFactory.create("pickle").load(path)

    @staticmethod
    def save(data, path):
        """
        Saves data to path.

        Args:
            data: data to save
            path: output path
        """

        # Save using default serialization method
        SerializeFactory.create().save(data, path)



"""
Transform module
"""

import numpy as np

from .action import Action


class Transform:
    """
    Executes a transform. Processes a stream of documents, loads batches into enabled data stores and vectorizes documents.
    """

    def __init__(self, embeddings, action, checkpoint=None):
        """
        Creates a new transform.

        Args:
            embeddings: embeddings instance
            action: index action
            checkpoint: optional checkpoint directory, enables indexing restart
        """

        self.embeddings = embeddings
        self.action = action
        self.checkpoint = checkpoint

        # Alias embeddings attributes
        self.config = embeddings.config
        self.delete = embeddings.delete
        self.model = embeddings.model
        self.database = embeddings.database
        self.graph = embeddings.graph
        self.indexes = embeddings.indexes
        self.scoring = embeddings.scoring if embeddings.issparse() else None

        # Get config parameters
        self.offset = embeddings.config.get("offset", 0) if action == Action.UPSERT else 0
        self.batch = embeddings.config.get("batch", 1024)

        # Scalar quantization
        quantize = embeddings.config.get("quantize")
        self.qbits = quantize if isinstance(quantize, int) and not isinstance(quantize, bool) else None

        # Transform columns
        columns = embeddings.config.get("columns", {})
        self.text = columns.get("text", "text")
        self.object = columns.get("object", "object")

        # Check if top-level indexing is enabled for this embeddings
        self.indexing = embeddings.model or embeddings.scoring

        # List of deleted ids with this action
        self.deletes = set()

    def __call__(self, documents, buffer):
        """
        Processes an iterable collection of documents, handles any iterable including generators.

        This method loads a stream of documents into enabled data stores and vectorizes documents into an embeddings array.

        Args:
            documents: iterable of (id, data, tags)
            buffer: file path used for memmap buffer

        Returns:
            (document ids, dimensions, embeddings)
        """

        # Return parameters
        ids, dimensions, embeddings = None, None, None

        if self.model:
            ids, dimensions, embeddings = self.vectors(documents, buffer)
        else:
            ids = self.ids(documents)

        return (ids, dimensions, embeddings)

    def vectors(self, documents, buffer):
        """
        Runs a vectors transform operation when dense indexing is enabled.

        Args:
            documents: iterable of (id, data, tags)
            buffer: file path used for memmap buffer

        Returns:
            (document ids, dimensions, embeddings)
        """

        # Determine dtype
        dtype = np.uint8 if self.qbits else np.float32

        # Transform documents into vectors
        return self.model.vectors(self.stream(documents), self.batch, self.checkpoint, buffer, dtype)

    def ids(self, documents):
        """
        Runs an ids transform operation when dense indexing is disabled.

        Args:
            documents: iterable of (id, data, tags)

        Returns:
            document ids
        """

        # Consume stream and build extract ids
        ids = []
        for uid, _, _ in self.stream(documents):
            ids.append(uid)

        # Save offset when dense indexing is disabled
        self.config["offset"] = self.offset

        return ids

    def stream(self, documents):
        """
        This method does two things:

        1. Filter and yield data to vectorize
        2. Batch and load original documents into enabled data stores (database, graph, scoring)

        Documents are yielded for vectorization if one of the following is True:
            - dict with a text or object field
            - not a dict

        Otherwise, documents are only batched and inserted into data stores

        Args:
            documents: iterable collection (id, data, tags)
        """

        # Batch and index offset. Index offset increments by count of documents streamed for vectorization
        batch, offset = [], 0

        # Iterate and process documents stream
        for document in documents:
            if isinstance(document[1], dict):
                # Set text field to uid when top-level indexing is disabled and text empty
                if not self.indexing and not document[1].get(self.text):
                    document[1][self.text] = str(document[0])

                if self.text in document[1]:
                    yield (document[0], document[1][self.text], document[2])
                    offset += 1
                elif self.object in document[1]:
                    yield (document[0], document[1][self.object], document[2])
                    offset += 1
            else:
                yield document
                offset += 1

            # Batch document
            batch.append(document)
            if len(batch) == self.batch:
                self.load(batch, offset)
                batch, offset = [], 0

        # Final batch
        if batch:
            self.load(batch, offset)

    def load(self, batch, offset):
        """
        Loads a document batch. This method deletes existing ids from an embeddings index and
        loads into enabled data stores (database, graph, scoring).

        Args:
            batch: list of (id, data, tags)
            offset: index offset for batch
        """

        # Delete from embeddings index first (which deletes from underlying indexes and datastores) if this is an upsert
        if self.action == Action.UPSERT:
            # Get list of ids not yet seen and deleted
            deletes = [uid for uid, _, _ in batch if uid not in self.deletes]
            if deletes:
                # Execute delete
                self.delete(deletes)

                # Save deleted ids as a delete must only occur once per action
                self.deletes.update(deletes)

        # Load batch into database except if this is a reindex
        if self.database and self.action != Action.REINDEX:
            self.database.insert(batch, self.offset)

        # Load batch into scoring
        if self.scoring:
            self.scoring.insert(batch, self.offset, self.checkpoint)

        # Load batch into subindex documents stream
        if self.indexes:
            self.indexes.insert(batch, self.offset, self.checkpoint)

        # Load batch into graph
        if self.graph:
            self.graph.insert(batch, self.offset)

        # Increment offset
        self.offset += offset



"""
Defines API paths for reranking endpoints.
"""

from typing import List, Optional

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/rerank")
def rerank(query: str, limit: Optional[int] = 3, factor: Optional[int] = 10):
    """
    Queries an embeddings database and reranks the results with a similarity pipeline.

    Args:
        query: query text
        limit: maximum results
        factor: factor to multiply limit by for the initial embeddings search

    Returns:
        query results
    """

    return application.get().pipeline("reranker", (query, limit, factor))


@router.post("/batchrerank")
def batchrerank(queries: List[str] = Body(...), limit: Optional[int] = Body(default=3), factor: Optional[int] = Body(default=10)):
    """
    Queries an embeddings database and reranks the results with a similarity pipeline.

    Args:
        queries: list of queries
        limit: maximum results
        factor: factor to multiply limit by for the initial embeddings search

    Returns:
        query results
    """

    return application.get().pipeline("reranker", (queries, limit, factor))



"""
PGVector module
"""

import os

import numpy as np

# Conditional import
try:
    from pgvector.sqlalchemy import BIT, HALFVEC, VECTOR

    from sqlalchemy import create_engine, delete, func, text, Column, Index, Integer, MetaData, StaticPool, Table
    from sqlalchemy.orm import Session
    from sqlalchemy.schema import CreateSchema

    PGVECTOR = True
except ImportError:
    PGVECTOR = False

from ..base import ANN


# pylint: disable=R0904
class PGVector(ANN):
    """
    Builds an ANN index backed by a Postgres database.
    """

    def __init__(self, config):
        super().__init__(config)

        if not PGVECTOR:
            raise ImportError('PGVector is not available - install "ann" extra to enable')

        # Database connection
        self.engine, self.database, self.connection, self.table = None, None, None, None

        # Scalar quantization
        quantize = self.config.get("quantize")
        self.qbits = quantize if quantize and isinstance(quantize, int) and not isinstance(quantize, bool) else None

    def load(self, path):
        # Initialize tables
        self.initialize()

    def index(self, embeddings):
        # Initialize tables
        self.initialize(recreate=True)

        # Prepare embeddings and insert rows
        self.database.execute(self.table.insert(), [{"indexid": x, "embedding": self.prepare(row)} for x, row in enumerate(embeddings)])

        # Create index
        self.createindex()

        # Add id offset and index build metadata
        self.config["offset"] = embeddings.shape[0]
        self.metadata(self.settings())

    def append(self, embeddings):
        # Prepare embeddings and insert rows
        self.database.execute(
            self.table.insert(), [{"indexid": x + self.config["offset"], "embedding": self.prepare(row)} for x, row in enumerate(embeddings)]
        )

        # Update id offset and index metadata
        self.config["offset"] += embeddings.shape[0]
        self.metadata()

    def delete(self, ids):
        self.database.execute(delete(self.table).where(self.table.c["indexid"].in_(ids)))

    def search(self, queries, limit):
        results = []
        for query in queries:
            # Run query
            query = self.database.query(self.table.c["indexid"], self.query(query)).order_by("score").limit(limit)

            # Calculate and collect scores
            results.append([(indexid, self.score(score)) for indexid, score in query])

        return results

    def count(self):
        # pylint: disable=E1102
        return self.database.query(func.count(self.table.c["indexid"])).scalar()

    def save(self, path):
        # Commit session and connection
        self.database.commit()
        self.connection.commit()

    def close(self):
        # Parent logic
        super().close()

        # Close database connection
        if self.database:
            self.database.close()
            self.engine.dispose()

    def initialize(self, recreate=False):
        """
        Initializes a new database session.

        Args:
            recreate: Recreates the database tables if True
        """

        # Connect to database
        self.connect()

        # Set the database schema
        self.schema()

        # Table name
        table = self.setting("table", self.defaulttable())

        # Create vectors table object
        self.table = Table(table, MetaData(), Column("indexid", Integer, primary_key=True, autoincrement=False), Column("embedding", self.column()))

        # Drop table, if necessary
        if recreate:
            self.table.drop(self.connection, checkfirst=True)

        # Create table, if necessary
        self.table.create(self.connection, checkfirst=True)

    def createindex(self):
        """
        Creates a index with the current settings.
        """

        # Table name
        table = self.setting("table", self.defaulttable())

        # Create ANN index - inner product is equal to cosine similarity on normalized vectors
        index = Index(
            f"{table}-index",
            self.table.c["embedding"],
            postgresql_using="hnsw",
            postgresql_with=self.settings(),
            postgresql_ops={"embedding": self.operation()},
        )

        # Create or recreate index
        index.drop(self.connection, checkfirst=True)
        index.create(self.connection, checkfirst=True)

    def connect(self):
        """
        Establishes a database connection. Cleans up any existing database connection first.
        """

        # Close existing connection
        if self.database:
            self.close()

        # Create engine
        self.engine = create_engine(self.url(), poolclass=StaticPool, echo=False)
        self.connection = self.engine.connect()

        # Start database session
        self.database = Session(self.connection)

        # Initialize pgvector extension
        self.sqldialect(text("CREATE EXTENSION IF NOT EXISTS vector"))

    def schema(self):
        """
        Sets the database schema, if available.
        """

        # Set default schema, if necessary
        schema = self.setting("schema")
        if schema:
            with self.engine.begin():
                self.sqldialect(CreateSchema(schema, if_not_exists=True))

            self.sqldialect(text("SET search_path TO :schema,public"), {"schema": schema})

    def settings(self):
        """
        Returns settings for this index.

        Returns:
            dict
        """

        return {"m": self.setting("m", 16), "ef_construction": self.setting("efconstruction", 200)}

    def sqldialect(self, sql, parameters=None):
        """
        Executes a SQL statement based on the current SQL dialect.

        Args:
            sql: SQL to execute
            parameters: optional bind parameters
        """

        args = (sql, parameters) if self.engine.dialect.name == "postgresql" else (text("SELECT 1"),)
        self.database.execute(*args)

    def defaulttable(self):
        """
        Returns the default table name.

        Returns:
            default table name
        """

        return "vectors"

    def url(self):
        """
        Reads the database url parameter.

        Returns:
            database url
        """

        return self.setting("url", os.environ.get("ANN_URL"))

    def column(self):
        """
        Gets embedding column for the current settings.

        Returns:
            embedding column definition
        """

        if self.qbits:
            # If quantization is set, always return BIT vectors
            return BIT(self.config["dimensions"] * 8)

        if self.setting("precision") == "half":
            # 16-bit HALF precision vectors
            return HALFVEC(self.config["dimensions"])

        # Default is full 32-bit FULL precision vectors
        return VECTOR(self.config["dimensions"])

    def operation(self):
        """
        Gets the index operation for the current settings.

        Returns:
            index operation
        """

        if self.qbits:
            # If quantization is set, always return BIT vectors
            return "bit_hamming_ops"

        if self.setting("precision") == "half":
            # 16-bit HALF precision vectors
            return "halfvec_ip_ops"

        # Default is full 32-bit FULL precision vectors
        return "vector_ip_ops"

    def prepare(self, data):
        """
        Prepares data for the embeddings column. This method returns a bit string for bit vectors and
        the input data unmodified for float vectors.

        Args:
            data: input data

        Returns:
            data ready for the embeddings column
        """

        # Transform to a bit string when vector quantization is enabled
        if self.qbits:
            return "".join(np.where(np.unpackbits(data), "1", "0"))

        # Return original data
        return data

    def query(self, query):
        """
        Creates a query statement from an input query. This method uses hamming distance for bit vectors and
        the max_inner_product for float vectors.

        Args:
            query: input query

        Returns:
            query statement
        """

        # Prepare query embeddings
        query = self.prepare(query)

        # Bit vector query
        if self.qbits:
            return self.table.c["embedding"].hamming_distance(query).label("score")

        # Float vector query
        return self.table.c["embedding"].max_inner_product(query).label("score")

    def score(self, score):
        """
        Calculates the index score from the input score. This method returns the hamming score
        (1.0 - (hamming distance / total number of bits)) for bit vectors and the -score for
        float vectors.

        Args:
            score: input score

        Returns:
            index score
        """

        # Calculate hamming score as 1.0 - (hamming distance / total number of bits)
        # Bound score from 0 to 1
        if self.qbits:
            return min(max(0.0, 1.0 - (score / (self.config["dimensions"] * 8))), 1.0)

        # pgvector returns negative inner product since Postgres only supports ASC order index scans on operators
        return -score



"""
Factory module
"""

from .json import JSONResponse
from .messagepack import MessagePackResponse


class ResponseFactory:
    """
    Methods to create Response classes.
    """

    @staticmethod
    def create(request):
        """
        Gets a response class for request using the Accept header.

        Args:
            request: request

        Returns:
            response class
        """

        # Get Accept header
        accept = request.headers.get("Accept")

        # Get response class
        return MessagePackResponse if accept == MessagePackResponse.media_type else JSONResponse



"""
Indexes module
"""

import os

from .documents import Documents


class Indexes:
    """
    Manages a collection of subindexes for an embeddings instance.
    """

    def __init__(self, embeddings, indexes):
        """
        Creates a new indexes instance.

        Args:
            embeddings: embeddings instance
            indexes: dict of subindexes to add
        """

        self.embeddings = embeddings
        self.indexes = indexes

        self.documents = None
        self.checkpoint = None

        # Transform columns
        columns = embeddings.config.get("columns", {})
        self.text = columns.get("text", "text")
        self.object = columns.get("object", "object")

        # Check if top-level indexing is enabled for this embeddings instance
        self.indexing = embeddings.model or embeddings.scoring

    def __contains__(self, name):
        """
        Returns True if name is in this instance, False otherwise.

        Returns:
            True if name is in this instance, False otherwise
        """

        return name in self.indexes

    def __getitem__(self, name):
        """
        Looks up an index by name.

        Args:
            name: index name

        Returns:
            index
        """

        return self.indexes[name]

    def __getattr__(self, name):
        """
        Looks up an index by attribute name.

        Args:
            name: index name

        Returns:
            index
        """

        try:
            return self.indexes[name]
        except Exception as e:
            raise AttributeError(e) from e

    def default(self):
        """
        Gets the default/first index.

        Returns:
            default index
        """

        return list(self.indexes.keys())[0]

    def findmodel(self, index=None):
        """
        Finds a vector model. If index is empty, the first vector model is returned.

        Args:
            index: index name to match

        Returns:
            Vectors
        """

        # Find vector model
        matches = [self.indexes[index].findmodel()] if index else [index.findmodel() for index in self.indexes.values() if index.findmodel()]
        return matches[0] if matches else None

    def insert(self, documents, index=None, checkpoint=None):
        """
        Inserts a batch of documents into each subindex.

        Args:
            documents: list of (id, data, tags)
            index: indexid offset
            checkpoint: optional checkpoint directory, enables indexing restart
        """

        if not self.documents:
            self.documents = Documents()
            self.checkpoint = checkpoint

        # Create batch containing documents added to parent index
        batch = []
        for _, document, _ in documents:
            # Add to documents collection if text or object field is set
            parent = document
            if isinstance(parent, dict):
                parent = parent.get(self.text, document.get(self.object))

            # Add if field is available or top-level indexing is disabled
            if parent is not None or not self.indexing:
                batch.append((index, document, None))
                index += 1

        # Add filtered documents batch
        self.documents.add(batch)

    def delete(self, ids):
        """
        Deletes ids from each subindex.

        Args:
            ids: list of ids to delete
        """

        for index in self.indexes.values():
            index.delete(ids)

    def index(self):
        """
        Builds each subindex.
        """

        for name, index in self.indexes.items():
            index.index(self.documents, checkpoint=f"{self.checkpoint}/{name}" if self.checkpoint else None)

        # Reset document stream
        self.documents.close()
        self.documents = None
        self.checkpoint = None

    def upsert(self):
        """
        Runs upsert for each subindex.
        """

        for index in self.indexes.values():
            index.upsert(self.documents)

        # Reset document stream
        self.documents.close()
        self.documents = None

    def load(self, path):
        """
        Loads each subindex from path.

        Args:
            path: directory path to load subindexes
        """

        for name, index in self.indexes.items():
            # Load subindex if it exists, subindexes aren't required to have data
            directory = os.path.join(path, name)
            if index.exists(directory):
                index.load(directory)

    def save(self, path):
        """
        Saves each subindex to path.

        Args:
            path: directory path to save subindexes
        """

        for name, index in self.indexes.items():
            index.save(os.path.join(path, name))

    def close(self):
        """
        Close and free resources used by this instance.
        """

        for index in self.indexes.values():
            index.close()



"""
Text imports
"""

from .crossencoder import CrossEncoder
from .entity import Entity
from .labels import Labels
from .lateencoder import LateEncoder
from .questions import Questions
from .reranker import Reranker
from .similarity import Similarity
from .summary import Summary
from .translation import Translation



"""
Factory module
"""

from ...util import Resolver

from .annoy import Annoy
from .faiss import Faiss
from .hnsw import HNSW
from .numpy import NumPy
from .pgvector import PGVector
from .sqlite import SQLite
from .torch import Torch


class ANNFactory:
    """
    Methods to create ANN indexes.
    """

    @staticmethod
    def create(config):
        """
        Create an ANN.

        Args:
            config: index configuration parameters

        Returns:
            ANN
        """

        # ANN instance
        ann = None
        backend = config.get("backend", "faiss")

        # Create ANN instance
        if backend == "annoy":
            ann = Annoy(config)
        elif backend == "faiss":
            ann = Faiss(config)
        elif backend == "hnsw":
            ann = HNSW(config)
        elif backend == "numpy":
            ann = NumPy(config)
        elif backend == "pgvector":
            ann = PGVector(config)
        elif backend == "sqlite":
            ann = SQLite(config)
        elif backend == "torch":
            ann = Torch(config)
        else:
            ann = ANNFactory.resolve(backend, config)

        # Store config back
        config["backend"] = backend

        return ann

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: index configuration parameters

        Returns:
            ANN
        """

        try:
            return Resolver()(backend)(config)
        except Exception as e:
            raise ImportError(f"Unable to resolve ann backend: '{backend}'") from e



"""
Cluster module
"""

import asyncio
import json
import random
import urllib.parse
import zlib

import aiohttp

from ..database.sql import Aggregate


class Cluster:
    """
    Aggregates multiple embeddings shards into a single logical embeddings instance.
    """

    # pylint: disable = W0231
    def __init__(self, config=None):
        """
        Creates a new Cluster.

        Args:
            config: cluster configuration
        """

        # Configuration
        self.config = config

        # Embeddings shard urls
        self.shards = None
        if "shards" in self.config:
            self.shards = self.config["shards"]

        # Query aggregator
        self.aggregate = Aggregate()

    def search(self, query, limit=None, weights=None, index=None, parameters=None, graph=False):
        """
        Finds documents most similar to the input query. This method will run either an index search
        or an index + database search depending on if a database is available.

        Args:
            query: input query
            limit: maximum results
            weights: hybrid score weights, if applicable
            index: index name, if applicable
            parameters: dict of named parameters to bind to placeholders
            graph: return graph results if True

        Returns:
            list of {id: value, score: value} for index search, list of dict for an index + database search
        """

        # Build URL
        action = f"search?query={urllib.parse.quote_plus(query)}"
        if limit:
            action += f"&limit={limit}"
        if weights:
            action += f"&weights={weights}"
        if index:
            action += f"&index={index}"
        if parameters:
            action += f"&parameters={json.dumps(parameters) if isinstance(parameters, dict) else parameters}"
        if graph is not None:
            action += f"&graph={graph}"

        # Run query and flatten results into single results list
        results = []
        for result in self.execute("get", action):
            results.extend(result)

        # Combine aggregate functions and sort
        results = self.aggregate(query, results)

        # Limit results
        return results[: (limit if limit else 10)]

    def batchsearch(self, queries, limit=None, weights=None, index=None, parameters=None, graph=False):
        """
        Finds documents most similar to the input queries. This method will run either an index search
        or an index + database search depending on if a database is available.

        Args:
            queries: input queries
            limit: maximum results
            weights: hybrid score weights, if applicable
            index: index name, if applicable
            parameters: list of dicts of named parameters to bind to placeholders
            graph: return graph results if True

        Returns:
            list of {id: value, score: value} per query for index search, list of dict per query for an index + database search
        """

        # POST parameters
        params = {"queries": queries}
        if limit:
            params["limit"] = limit
        if weights:
            params["weights"] = weights
        if index:
            params["index"] = index
        if parameters:
            params["parameters"] = parameters
        if graph is not None:
            params["graph"] = graph

        # Run query
        batch = self.execute("post", "batchsearch", [params] * len(self.shards))

        # Combine results per query
        results = []
        for x, query in enumerate(queries):
            result = []
            for section in batch:
                result.extend(section[x])

            # Aggregate, sort and limit results
            results.append(self.aggregate(query, result)[: (limit if limit else 10)])

        return results

    def add(self, documents):
        """
        Adds a batch of documents for indexing.

        Args:
            documents: list of {id: value, text: value}
        """

        self.execute("post", "add", self.shard(documents))

    def index(self):
        """
        Builds an embeddings index for previously batched documents.
        """

        self.execute("get", "index")

    def upsert(self):
        """
        Runs an embeddings upsert operation for previously batched documents.
        """

        self.execute("get", "upsert")

    def delete(self, ids):
        """
        Deletes from an embeddings cluster. Returns list of ids deleted.

        Args:
            ids: list of ids to delete

        Returns:
            ids deleted
        """

        return [uid for ids in self.execute("post", "delete", [ids] * len(self.shards)) for uid in ids]

    def reindex(self, config, function=None):
        """
        Recreates this embeddings index using config. This method only works if document content storage is enabled.

        Args:
            config: new config
            function: optional function to prepare content for indexing
        """

        self.execute("post", "reindex", [{"config": config, "function": function}] * len(self.shards))

    def count(self):
        """
        Total number of elements in this embeddings cluster.

        Returns:
            number of elements in embeddings cluster
        """

        return sum(self.execute("get", "count"))

    def shard(self, documents):
        """
        Splits documents into equal sized shards.

        Args:
            documents: input documents

        Returns:
            list of evenly sized shards with the last shard having the remaining elements
        """

        shards = [[] for _ in range(len(self.shards))]
        for document in documents:
            uid = document.get("id") if isinstance(document, dict) else document
            if uid and isinstance(uid, str):
                # Quick int hash of string to help derive shard id
                uid = zlib.adler32(uid.encode("utf-8"))
            elif uid is None:
                # Get random shard id when uid isn't set
                uid = random.randint(0, len(shards) - 1)

            shards[uid % len(self.shards)].append(document)

        return shards

    def execute(self, method, action, data=None):
        """
        Executes a HTTP action asynchronously.

        Args:
            method: get or post
            action: url action to perform
            data: post parameters

        Returns:
            json results if any
        """

        # Get urls
        urls = [f"{shard}/{action}" for shard in self.shards]
        close = False

        # Use existing loop if available, otherwise create one
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            close = True

        try:
            return loop.run_until_complete(self.run(urls, method, data))
        finally:
            # Close loop if it was created in this method
            if close:
                loop.close()

    async def run(self, urls, method, data):
        """
        Runs an async action.

        Args:
            urls: run against this list of urls
            method: get or post
            data: list of data for each url or None

        Returns:
            json results if any
        """

        async with aiohttp.ClientSession(raise_for_status=True) as session:
            tasks = []

            for x, url in enumerate(urls):
                if method == "post":
                    if not data or data[x]:
                        tasks.append(asyncio.ensure_future(self.post(session, url, data[x] if data else None)))
                else:
                    tasks.append(asyncio.ensure_future(self.get(session, url)))

            return await asyncio.gather(*tasks)

    async def get(self, session, url):
        """
        Runs an async HTTP GET request.

        Args:
            session: ClientSession
            url: url

        Returns:
            json results if any
        """

        async with session.get(url) as resp:
            return await resp.json()

    async def post(self, session, url, data):
        """
        Runs an async HTTP POST request.

        Args:
            session: ClientSession
            url: url
            data: data to POST

        Returns:
            json results if any
        """

        async with session.post(url, json=data) as resp:
            return await resp.json()
